"""
Class wrappers for the Key/Value backend
"""

import logging

from saltext.vault.utils.vault.exceptions import VaultException
from saltext.vault.utils.vault.exceptions import VaultInvocationError
from saltext.vault.utils.vault.exceptions import VaultNotFoundError
from saltext.vault.utils.vault.exceptions import VaultPermissionDeniedError
from saltext.vault.utils.vault.exceptions import VaultUnsupportedOperationError

log = logging.getLogger(__name__)


class VaultKV:
    """
    Interface to Vault secret paths
    """

    def __init__(self, client, metadata_cache):
        self.client = client
        self.metadata_cache = metadata_cache

    def read(self, path, include_metadata=False, version=None):
        """
        Read secret data at path.

        include_metadata
            For KV v2, include metadata in the return value:
            ``{"data": {} ,"metadata": {}}``.
        """
        v2_info = self.is_v2(path)
        if v2_info["v2"]:
            path = v2_info["data"]
        elif version is not None:
            if version != 0:
                raise VaultInvocationError("Cannot request secret versions on KV v1")
            version = None
        if version is not None:
            path += f"?version={version}"
        res = self.client.get(path)
        ret = res["data"]
        if v2_info["v2"] and not include_metadata:
            return ret["data"]
        return ret

    def read_meta(self, path):
        """
        Read secret metadata for all versions at path. This is different from
        the metadata returned by read, which pertains only to the most recent
        version. Requires KV v2.

        .. versionadded:: 1.2.0
        """
        v2_info = self.is_v2(path)
        if not v2_info["v2"]:
            raise VaultInvocationError("The backend is not KV v2")
        return self.client.get(v2_info["metadata"])["data"]

    def write(self, path, data, cas=None):
        """
        Write secret data to path.
        """
        v2_info = self.is_v2(path)
        if v2_info["v2"]:
            path = v2_info["data"]
            data = {"data": data}
            if cas is not None:
                data["options"] = {"cas": cas}
        # Always treat this request as idempotent, even though it's technically not
        # on KVv2 (writing the same data twice increases the current version twice).
        return self.client.put(path, payload=data)

    def patch(self, path, data):
        """
        Patch existing data.
        Tries to use a PATCH request, otherwise falls back to updating in memory
        and writing back the whole secret, thus might consume more than one token use.

        Since this uses the `JSON Merge Patch format <https://datatracker.ietf.org/doc/html/draft-ietf-appsawg-json-merge-patch-07>`_,
        values set to ``null`` (``None``) will be dropped.
        """

        def apply_json_merge_patch(data, patch):
            if not patch:
                return data
            if not isinstance(data, dict) or not isinstance(patch, dict):
                raise ValueError("Data and patch must be dictionaries.")

            for key, value in patch.items():
                if value is None:
                    data.pop(key, None)
                elif isinstance(value, dict):
                    data[key] = apply_json_merge_patch(data.get(key, {}), value)
                else:
                    data[key] = value
            return data

        v2_info = self.is_v2(path)

        def patch_in_memory(path, data):
            current = self.read(path, include_metadata=v2_info["v2"])
            updated = apply_json_merge_patch(current["data"] if v2_info["v2"] else current, data)
            return self.write(
                path, updated, cas=current["metadata"]["version"] if v2_info["v2"] else None
            )

        if not v2_info["v2"]:
            return patch_in_memory(path, data)

        path = v2_info["data"]
        payload = {"data": data}
        try:
            return self.client.patch(path, payload=payload, safe_to_retry=True)
        except VaultPermissionDeniedError:
            log.warning("Failed patching secret, is the `patch` capability set?")
        except VaultUnsupportedOperationError:
            pass
        return patch_in_memory(path, data)

    def delete(self, path, versions=None, all_versions=False):
        """
        Delete secret path data. For KV v1, this is permanent.
        For KV v2, this only soft-deletes the data.

        versions
            For KV v2, specifies versions to soft-delete. Needs to be castable
            to a list of integers.

        all_versions
            For KV v2, delete all known versions. Defaults to false.

            .. versionadded:: 1.2.0

        """
        method = "DELETE"
        payload = None
        v2_info = self.is_v2(path)
        safe_to_retry = None

        if all_versions and v2_info["v2"]:
            versions = []
            try:
                curr = self.read_meta(path)
            except VaultNotFoundError:
                # The delete API behaves the same
                return True
            for version, meta in curr["versions"].items():
                if not meta["destroyed"] and not meta["deletion_time"]:
                    versions.append(version)
            if not versions:
                # No version left to delete
                return True
        versions = self._parse_versions(versions)

        if v2_info["v2"]:
            if versions is not None:
                method = "POST"
                path = v2_info["delete_versions"]
                payload = {"versions": versions}
                safe_to_retry = True
            else:
                # data and delete operations only differ by HTTP verb
                path = v2_info["data"]
        elif versions is not None:
            raise VaultInvocationError("Versioning support requires KV v2.")

        return self.client.request(method, path, payload=payload, safe_to_retry=safe_to_retry)

    def restore(self, path, versions=None, all_versions=False):
        """
        .. versionadded:: 1.2.0

        Restore secret versions. Requires KV v2.

        versions
            Specifies soft-deleted versions of a secret path
            to restore. Needs to be castable to a list of integers.
            If unspecified and the latest version of a secret
            is deleted, restores this version, otherwise fails.

        all_versions
            Restore all soft-deleted versions of the secret. Defaults to false.
        """
        v2_info = self.is_v2(path)
        if not v2_info["v2"]:
            raise VaultInvocationError("Destroy operation requires KV v2.")
        if all_versions or not versions:
            versions = []
            curr = self.read_meta(path)["versions"]
            if all_versions:
                for version, meta in curr.items():
                    if not meta["destroyed"] and meta["deletion_time"]:
                        versions.append(version)
            else:
                most_recent = str(max(int(x) for x in curr))
                if not curr[most_recent]["destroyed"] and curr[most_recent]["deletion_time"]:
                    versions = [most_recent]
            if not versions:
                # No version left to destroy
                raise VaultInvocationError("No secret version to restore.")

        versions = self._parse_versions(versions)
        path = v2_info["undelete"]
        payload = {"versions": versions}
        return self.client.post(path, payload=payload)

    def destroy(self, path, versions=None, all_versions=False):
        """
        Permanently remove version data. Requires KV v2.

        versions
            Specifies versions to destroy. Needs to be castable
            to a list of integers.

            .. versionchanged:: 1.2.0
                If unspecified, destroys the most recent version.

        all_versions
            Destroy all versions of the secret. Defaults to false.

            .. versionadded:: 1.2.0
        """
        v2_info = self.is_v2(path)
        if not v2_info["v2"]:
            raise VaultInvocationError("Destroy operation requires KV v2.")
        if all_versions or not versions:
            versions = []
            try:
                curr = self.read_meta(path)["versions"]
            except VaultNotFoundError:
                # The destroy API behaves the same
                return True
            if all_versions:
                for version, meta in curr.items():
                    if not meta["destroyed"]:
                        versions.append(version)
            else:
                most_recent = str(max(int(x) for x in curr))
                if not curr[most_recent]["destroyed"]:
                    versions = [most_recent]
            if not versions:
                # No version left to destroy
                return True

        versions = self._parse_versions(versions)
        path = v2_info["destroy"]
        payload = {"versions": versions}
        return self.client.post(path, payload=payload)

    def _parse_versions(self, versions):
        if versions is None:
            return versions
        if not isinstance(versions, list):
            versions = [versions]
        try:
            versions = [int(x) for x in versions]
        except ValueError as err:
            raise VaultInvocationError("Versions have to be specified as integers.") from err
        return versions

    def nuke(self, path):
        """
        Delete path metadata and version data, including all version history.
        Requires KV v2.
        """
        v2_info = self.is_v2(path)
        if not v2_info["v2"]:
            raise VaultInvocationError("Wipe operation requires KV v2.")
        path = v2_info["metadata"]
        return self.client.delete(path)

    def list(self, path):
        """
        List keys at path.
        """
        v2_info = self.is_v2(path)
        if v2_info["v2"]:
            path = v2_info["metadata"]

        return self.client.list(path)["data"]["keys"]

    def is_v2(self, path):
        """
        Determines if a given secret path is KV v1 or v2.
        """
        ret = {
            "v2": False,
            "data": path,
            "metadata": path,
            "delete": path,
            "type": None,
        }
        path_metadata = self._get_secret_path_metadata(path)
        if not path_metadata:
            # metadata lookup failed. Simply return not v2
            return ret
        ret["type"] = path_metadata.get("type", "kv")
        if (
            ret["type"] == "kv"
            and path_metadata["options"] is not None
            and path_metadata.get("options", {}).get("version", "1") == "2"
        ):
            ret["v2"] = True
            ret["data"] = self._v2_the_path(path, path_metadata.get("path", path))
            ret["metadata"] = self._v2_the_path(path, path_metadata.get("path", path), "metadata")
            ret["delete"] = ret["data"]
            ret["delete_versions"] = self._v2_the_path(
                path, path_metadata.get("path", path), "delete"
            )
            ret["destroy"] = self._v2_the_path(path, path_metadata.get("path", path), "destroy")
            ret["undelete"] = self._v2_the_path(path, path_metadata.get("path", path), "undelete")
        return ret

    def _v2_the_path(self, path, pfilter, ptype="data"):
        """
        Given a path, a filter, and a path type, properly inject
        'data' or 'metadata' into the path.
        """
        possible_types = ("data", "metadata", "delete", "destroy", "undelete")
        if ptype not in possible_types:
            raise AssertionError()
        msg = f"Path {path} already contains {ptype} in the right place - saltstack duct tape?"

        path = path.rstrip("/").lstrip("/")
        pfilter = pfilter.rstrip("/").lstrip("/")

        together = pfilter + "/" + ptype

        otype = possible_types[0] if possible_types[0] != ptype else possible_types[1]
        other = pfilter + "/" + otype
        if path.startswith(other):
            path = path.replace(other, together, 1)
            msg = f'Path is a "{otype}" type but "{ptype}" type requested - Flipping: {path}'
        elif not path.startswith(together):
            old_path = path
            path = path.replace(pfilter, together, 1)
            msg = f"Converting path to v2 {old_path} => {path}"
        log.debug(msg)
        return path

    def _get_secret_path_metadata(self, path):
        """
        Given a path, query vault to determine mount point, type, and version.
        """
        cache_content = self.metadata_cache.get() or {}

        ret = None
        if path.startswith(tuple(cache_content.keys())):
            log.debug("Found cached metadata for %s", path)
            ret = next(v for k, v in cache_content.items() if path.startswith(k))
        else:
            log.debug("Fetching metadata for %s", path)
            try:
                endpoint = f"sys/internal/ui/mounts/{path}"
                res = self.client.get(endpoint)
                if "data" in res:
                    log.debug("Got metadata for %s", path)
                    cache_content[path] = ret = res["data"]
                    self.metadata_cache.store(cache_content)
                else:
                    raise VaultException("Unexpected response to metadata query.")
            except Exception as err:  # pylint: disable=broad-except
                log.error("Failed to get secret metadata %s: %s", type(err).__name__, err)
        return ret



import pytest

from saltext.vault.utils.vault import db


@pytest.mark.parametrize(
    "name,expected",
    (
        ("mysql", {"name": "mysql", "required": ["connection_url"]}),
        ("custom", {"name": "", "required": []}),
    ),
)
def test_get_plugin_meta(name, expected):
    ret = db.get_plugin_meta(name).copy()
    assert ret == expected


@pytest.mark.parametrize(
    "name,expected",
    (
        ("mysql", "mysql-database-plugin"),
        ("redis_elasticache", "redis-elasticache-database-plugin"),
        ("custom", "custom-database-plugin"),
    ),
)
def test_get_plugin_name(name, expected):
    ret = db.get_plugin_name(name)
    assert ret == expected


@pytest.mark.parametrize(
    "name,mount,cache,static,expected",
    (
        (None, None, None, None, "db.*.*.*.*"),
        ("test", None, None, None, "db.*.*.test.*"),
        ("test", None, True, None, "db.*.*.test.default"),
        ("test", None, "alt", None, "db.*.*.test.alt"),
        ("test", None, "alt", False, "db.*.dynamic.test.alt"),
        ("test", None, "alt", True, "db.*.static.test.default"),
        ("test", None, None, True, "db.*.static.test.default"),
        ("test", None, True, True, "db.*.static.test.default"),
        ("test", "foo", True, True, "db.foo.static.test.default"),
        ("test", "foo", None, True, "db.foo.static.test.default"),
        ("test", "foo", None, False, "db.foo.dynamic.test.*"),
        ("test", "foo", "alt2", False, "db.foo.dynamic.test.alt2"),
    ),
)
def test_create_cache_pattern(name, mount, cache, static, expected):
    ret = db.create_cache_pattern(name=name, mount=mount, cache=cache, static=static)
    assert ret == expected



"""
Vault authentication models
"""

import logging

from saltext.vault.utils.vault import leases
from saltext.vault.utils.vault.exceptions import VaultAuthExpired

log = logging.getLogger(__name__)


class VaultTokenAuth:
    """
    Container for authentication tokens
    """

    def __init__(self, cache=None, token=None):
        self.cache = cache
        if token is None and cache is not None:
            token = cache.get()
        if token is None:
            token = InvalidVaultToken()
        if isinstance(token, dict):
            token = leases.VaultToken(**token)
        self.token = token

    def is_renewable(self):
        """
        Check whether the contained token is renewable, which requires it
        to be currently valid for at least two uses and renewable
        """
        return self.token.is_renewable()

    def is_valid(self, valid_for=0):
        """
        Check whether the contained token is valid
        """
        return self.token.is_valid(valid_for)

    def get_token(self):
        """
        Get the contained token if it is valid, otherwise
        raises VaultAuthExpired
        """
        if self.token.is_valid():
            return self.token
        raise VaultAuthExpired()

    def used(self):
        """
        Increment the use counter for the contained token
        """
        self.token.used()
        if self.token.num_uses != 0:
            self._write_cache()

    def update_token(self, auth):
        """
        Partially update the contained token (e.g. after renewal)
        """
        self.token = self.token.with_renewed(**auth)
        self._write_cache()

    def replace_token(self, token):
        """
        Completely replace the contained token with a new one
        """
        self.token = token
        self._write_cache()

    def _write_cache(self):
        if self.cache is not None:
            # Write the token indiscriminately since flushing
            # raises VaultAuthExpired.
            # This will be handled as part of the next request.
            self.cache.store(self.token)


class VaultAppRoleAuth:
    """
    Issues tokens from AppRole credentials.
    """

    def __init__(self, approle, client, mount="approle", cache=None, token_store=None):
        self.approle = approle
        self.client = client
        self.mount = mount
        self.cache = cache
        if token_store is None:
            token_store = VaultTokenAuth()
        self.token = token_store

    def is_renewable(self):
        """
        Check whether the currently used token is renewable.
        SecretIDs are not renewable anyways.
        """
        return self.token.is_renewable()

    def is_valid(self, valid_for=0):
        """
        Check whether the contained authentication data can be used
        to issue a valid token
        """
        return self.token.is_valid(valid_for) or self.approle.is_valid(valid_for)

    def get_token(self):
        """
        Return the token issued by the last login, if it is still valid, otherwise
        login with the contained AppRole, if it is valid. Otherwise,
        raises VaultAuthExpired
        """
        if self.token.is_valid():
            return self.token.get_token()
        if self.approle.is_valid():
            return self._login()
        raise VaultAuthExpired()

    def used(self):
        """
        Increment the use counter for the currently used token
        """
        self.token.used()

    def update_token(self, auth):
        """
        Partially update the contained token (e.g. after renewal)
        """
        self.token.update_token(auth)

    def _login(self):
        log.debug("Vault token expired. Recreating one by authenticating with AppRole.")
        endpoint = f"auth/{self.mount}/login"
        payload = self.approle.payload()
        res = self.client.post(endpoint, payload=payload)
        self.approle.used()
        self._replace_token(res["auth"])
        self._write_cache()
        return self.token.get_token()

    def _write_cache(self):
        if self.cache is not None and self.approle.secret_id is not None:
            if isinstance(self.approle.secret_id, LocalVaultSecretId):
                pass
            elif self.approle.secret_id.num_uses == 0:
                pass
            elif self.approle.secret_id.is_valid():
                self.cache.store(self.approle.secret_id)
            else:
                self.cache.flush()

    def _replace_token(self, auth):
        self.token.replace_token(leases.VaultToken(**auth))


class VaultAppRole:
    """
    Container that represents an AppRole
    """

    def __init__(self, role_id, secret_id=None):
        self.role_id = role_id
        self.secret_id = secret_id

    def replace_secret_id(self, secret_id):
        """
        Replace the contained SecretID with a new one
        """
        self.secret_id = secret_id

    def is_valid(self, valid_for=0, uses=1):
        """
        Checks whether the contained data can be used to authenticate
        to Vault. SecretIDs might not be required by the server when
        bind_secret_id is set to false.

        valid_for
            Allows to check whether the AppRole will still be valid in the future.
            This can be an integer, which will be interpreted as seconds, or a
            time string using the same format as Vault does:
            Suffix ``s`` for seconds, ``m`` for minutes, ``h`` for hours, ``d`` for days.
            Defaults to 0.

        uses
            Check whether the AppRole has at least this number of uses left. Defaults to 1.
        """
        if self.secret_id is None:
            return True
        return self.secret_id.is_valid(valid_for=valid_for, uses=uses)

    def used(self):
        """
        Increment the SecretID use counter by one, if this AppRole uses one.
        """
        if self.secret_id is not None:
            self.secret_id.used()

    def payload(self):
        """
        Return the payload to use for POST requests using this AppRole
        """
        payload = {}
        if self.secret_id is not None:
            payload = self.secret_id.payload()
        payload["role_id"] = self.role_id
        return payload


class LocalVaultSecretId(leases.VaultSecretId):
    """
    Represents a SecretID from local configuration and should not be cached.
    """

    def is_valid(self, valid_for=0, uses=1):  # pylint: disable=unused-argument
        """
        Local SecretIDs are always assumed to be valid until proven otherwise
        """
        return True


class InvalidVaultToken(leases.VaultToken):
    """
    Represents a missing token
    """

    def __init__(self, *args, **kwargs):  # pylint: disable=super-init-not-called,unused-argument
        self.renewable = False
        self.use_count = 0
        self.num_uses = 0

    def is_valid(self, valid_for=0, uses=1):  # pylint: disable=unused-argument
        return False


class InvalidVaultSecretId(leases.VaultSecretId):
    """
    Represents a missing SecretID
    """

    def __init__(self, *args, **kwargs):  # pylint: disable=super-init-not-called
        pass

    def is_valid(self, valid_for=0, uses=1):  # pylint: disable=unused-argument
        return False



"""
Allows to link configuration values using :vconf:`config_val`.
"""


def setup(app):
    app.add_crossref_type(
        directivename="vconf",
        rolename="vconf",
        indextemplate="single: configuration; %s",
    )
    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""
Statefully manage Vault (or OpenBao) policies.

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.
"""

import difflib
import logging

from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def policy_present(name, rules):
    """
    Ensure a Vault policy with the given name and rules is present.

    name
        The name of the policy

    rules
        Rules formatted as in-line HCL


    .. code-block:: yaml

        demo-policy:
          vault.policy_present:
            - name: foo/bar
            - rules: |
                path "secret/top-secret/*" {
                  policy = "deny"
                }
                path "secret/not-very-secret/*" {
                  policy = "write"
                }

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    try:
        existing_rules = __salt__["vault.policy_fetch"](name)
    except CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = f"Failed to read policy: {err}"
        return ret

    if existing_rules == rules:
        ret["comment"] = "Policy exists, and has the correct content"
        return ret

    diff = "".join(
        difflib.unified_diff((existing_rules or "").splitlines(True), rules.splitlines(True))
    )

    ret["changes"] = {name: diff}

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Policy would be " + ("created" if existing_rules is None else "updated")
        return ret

    try:
        __salt__["vault.policy_write"](name, rules)
        ret["comment"] = "Policy has been " + ("created" if existing_rules is None else "updated")
        return ret
    except CommandExecutionError as err:
        return {
            "name": name,
            "changes": {},
            "result": False,
            "comment": f"Failed to write policy: {err}",
        }


def policy_absent(name):
    """
    Ensure a Vault policy with the given name and rules is absent.

    name
        The name of the policy
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    try:
        existing_rules = __salt__["vault.policy_fetch"](name)
    except CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = f"Failed to read policy: {err}"
        return ret

    if existing_rules is None:
        ret["comment"] = "Policy is already absent"
        return ret

    ret["changes"] = {"deleted": name}

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Policy would be deleted"
        return ret

    try:
        if not __salt__["vault.policy_delete"](name):
            raise CommandExecutionError(
                "Policy was initially reported as existent, but seemed to be "
                "absent while deleting."
            )
        ret["comment"] = "Policy has been deleted"
        return ret
    except CommandExecutionError as err:
        return {
            "name": name,
            "changes": {},
            "result": False,
            "comment": f"Failed to delete policy: {err}",
        }



import contextlib
from pathlib import Path
from unittest.mock import ANY
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch

import pytest
import salt.exceptions

import saltext.vault.utils.vault as vaultutil
from saltext.vault.runners import vault
from saltext.vault.utils.vault import api as vapi
from saltext.vault.utils.vault import client as vclient


@pytest.fixture
def configure_loader_modules(master_opts):
    return {
        vault: {
            "__grains__": {"id": "test-master"},
            "__opts__": master_opts,
        }
    }


@pytest.fixture
def default_config():
    return {
        "auth": {
            "approle_mount": "approle",
            "approle_name": "salt-master",
            "method": "token",
            "token": "test-token",
            "role_id": "test-role-id",
            "secret_id": None,
            "token_lifecycle": {
                "minimum_ttl": 10,
                "renew_increment": None,
            },
        },
        "cache": {
            "backend": "session",
            "config": 3600,
            "kv_metadata": "connection",
            "secret": "ttl",
        },
        "client": {
            "connect_timeout": vclient.DEFAULT_CONNECT_TIMEOUT,
            "read_timeout": vclient.DEFAULT_READ_TIMEOUT,
            "max_retries": vclient.DEFAULT_MAX_RETRIES,
            "backoff_factor": vclient.DEFAULT_BACKOFF_FACTOR,
            "backoff_max": vclient.DEFAULT_BACKOFF_MAX,
            "backoff_jitter": vclient.DEFAULT_BACKOFF_JITTER,
            "retry_post": vclient.DEFAULT_RETRY_POST,
            "retry_status": vclient.DEFAULT_RETRY_STATUS,
            "respect_retry_after": vclient.DEFAULT_RESPECT_RETRY_AFTER,
            "retry_after_max": vclient.DEFAULT_RETRY_AFTER_MAX,
        },
        "issue": {
            "allow_minion_override_params": False,
            "block_minion_requests": False,
            "type": "token",
            "approle": {
                "mount": "salt-minions",
                "params": {
                    "bind_secret_id": True,
                    "secret_id_num_uses": 1,
                    "secret_id_ttl": 60,
                    "token_explicit_max_ttl": 9999999999,
                    "token_num_uses": 1,
                },
            },
            "token": {
                "role_name": None,
                "params": {
                    "explicit_max_ttl": 9999999999,
                    "num_uses": 1,
                },
            },
            "wrap": "30s",
        },
        "issue_params": {},
        "metadata": {
            "entity": {
                "minion-id": "{minion}",
            },
            "secret": {
                "saltstack-jid": "{jid}",
                "saltstack-minion": "{minion}",
                "saltstack-user": "{user}",
            },
        },
        "policies": {
            "assign": [
                "saltstack/minions",
                "saltstack/{minion}",
            ],
            "cache_time": 60,
            "refresh_pillar": None,
        },
        "server": {
            "url": "http://test-vault:8200",
            "namespace": None,
            "verify": None,
        },
    }


@pytest.fixture
def token_response():
    return {
        "request_id": "0e8c388e-2cb6-bcb2-83b7-625127d568bb",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "auth": {
            "client_token": "test-token",
            "renewable": True,
            "lease_duration": 9999999999,
            "num_uses": 1,
            "creation_time": 1661188581,
            # "expire_time": 11661188580,
        },
    }


@pytest.fixture
def secret_id_response():
    return {
        "request_id": "0e8c388e-2cb6-bcb2-83b7-625127d568bb",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": {
            "secret_id_accessor": "84896a0c-1347-aa90-a4f6-aca8b7558780",
            "secret_id": "841771dc-11c9-bbc7-bcac-6a3945a69cd9",
            "secret_id_ttl": 60,
        },
    }


@pytest.fixture
def wrapped_response():
    return {
        "request_id": "",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": None,
        "warnings": None,
        "wrap_info": {
            "token": "test-wrapping-token",
            "accessor": "test-wrapping-token-accessor",
            "ttl": 180,
            "creation_time": "2022-09-10T13:37:12.123456789+00:00",
            "creation_path": "whatever/not/checked/here",
            "wrapped_accessor": "84896a0c-1347-aa90-a4f6-aca8b7558780",
        },
    }


@pytest.fixture
def token_serialized(token_response):
    return {
        "client_token": token_response["auth"]["client_token"],
        "renewable": token_response["auth"]["renewable"],
        "lease_duration": token_response["auth"]["lease_duration"],
        "num_uses": token_response["auth"]["num_uses"],
        "creation_time": token_response["auth"]["creation_time"],
        # "expire_time": token_response["auth"]["expire_time"],
    }


@pytest.fixture
def secret_id_serialized(secret_id_response):
    return {
        "secret_id": secret_id_response["data"]["secret_id"],
        "secret_id_ttl": secret_id_response["data"]["secret_id_ttl"],
        "secret_id_num_uses": 1,
        # + creation_time
        # + expire_time
    }


@pytest.fixture
def wrapped_serialized(wrapped_response):
    return {
        "wrap_info": {
            "token": wrapped_response["wrap_info"]["token"],
            "ttl": wrapped_response["wrap_info"]["ttl"],
            "creation_time": 1662817032,
            "creation_path": wrapped_response["wrap_info"]["creation_path"],
        },
    }


@pytest.fixture
def approle_meta(token_serialized, secret_id_serialized):
    return {
        "bind_secret_id": True,
        "local_secret_ids": False,
        "secret_id_bound_cidrs": [],
        "secret_id_num_uses": secret_id_serialized["secret_id_num_uses"],
        "secret_id_ttl": secret_id_serialized["secret_id_ttl"],
        "token_bound_cidrs": [],
        "token_explicit_max_ttl": token_serialized["lease_duration"],
        "token_max_ttl": 0,
        "token_no_default_policy": False,
        "token_num_uses": token_serialized["num_uses"],
        "token_period": 0,
        "token_policies": ["default"],
        "token_ttl": 0,
        "token_type": "default",
    }


@pytest.fixture
def policies_default():
    return ["saltstack/minions", "saltstack/minion/test-minion"]


@pytest.fixture
def metadata_secret_default():
    return {
        "saltstack-jid": "<no jid set>",
        "saltstack-minion": "test-minion",
        "saltstack-user": "<no user set>",
    }


@pytest.fixture
def metadata_entity_default():
    return {"minion-id": "test-minion"}


@pytest.fixture
def grains():
    return {
        "id": "test-minion",
        "roles": ["web", "database"],
        "aux": ["foo", "bar"],
        "deep": {"foo": {"bar": {"baz": ["hello", "world"]}}},
        "mixedcase": "UP-low-UP",
    }


@pytest.fixture
def pillar():
    return {
        "mixedcase": "UP-low-UP",
        "role": "test",
        "roles": ["foo", "bar"],
    }


@pytest.fixture
def client():
    with patch("saltext.vault.runners.vault._get_master_client", autospec=True) as get_client:
        client = Mock(spec=vclient.AuthenticatedVaultClient)
        get_client.return_value = client
        yield client


@pytest.fixture
def approle_api():
    with patch("saltext.vault.runners.vault._get_approle_api", autospec=True) as get_api:
        api = Mock(spec=vapi.AppRoleApi)
        get_api.return_value = api
        yield api


@pytest.fixture
def identity_api():
    with patch("saltext.vault.runners.vault._get_identity_api", autospec=True) as get_api:
        api = Mock(spec=vapi.IdentityApi)
        get_api.return_value = api
        yield api


@pytest.fixture
def client_token(client, token_response, wrapped_response):  # pylint: disable=unused-argument
    def res_or_wrap(*args, **kwargs):  # pylint: disable=unused-argument
        if kwargs.get("wrap"):
            return vaultutil.VaultWrappedResponse(**wrapped_response["wrap_info"])
        return token_response

    client.post.side_effect = res_or_wrap
    yield client


@pytest.fixture
def config(request, default_config):
    def rec(  # pylint: disable=inconsistent-return-statements, duplicate-code
        config, path, val=None, default=vaultutil.VaultException
    ):
        ptr = config
        parts = path.split(":")
        while parts:
            cur = parts.pop(0)
            if val:
                if parts and not isinstance(ptr.get(cur), dict):
                    ptr[cur] = {}
                elif not parts:
                    ptr[cur] = val
                    return
            if cur not in ptr:
                if isinstance(default, Exception):
                    raise default()
                return default
            ptr = ptr[cur]
        return ptr

    def get_config(key=None, default=vaultutil.VaultException):
        overrides = getattr(request, "param", {})
        if key is None:
            for ovar, oval in overrides.items():
                rec(default_config, ovar, oval)
            return default_config
        if key in overrides:
            return overrides[key]
        return rec(default_config, key, default=default)

    with patch("saltext.vault.runners.vault._config", autospec=True) as config:
        config.side_effect = get_config
        yield config


@pytest.fixture
def policies(request, policies_default):
    policies_list = getattr(request, "param", policies_default)
    with patch(
        "saltext.vault.runners.vault._get_policies_cached", autospec=True
    ) as get_policies_cached:
        get_policies_cached.return_value = policies_list
        with patch("saltext.vault.runners.vault._get_policies", autospec=True) as get_policies:
            get_policies.return_value = policies_list
            yield


@pytest.fixture
def metadata(request, metadata_entity_default, metadata_secret_default):
    def _get_metadata(
        minion_id, metadata_patterns, *args, **kwargs
    ):  # pylint: disable=unused-argument
        if getattr(request, "param", None) is not None:
            return request.param
        if "saltstack-jid" not in metadata_patterns:
            return metadata_entity_default
        return metadata_secret_default

    with patch("saltext.vault.runners.vault._get_metadata", autospec=True) as get_metadata:
        get_metadata.side_effect = _get_metadata
        yield get_metadata


@pytest.fixture
def validate_signature():
    with patch(
        "saltext.vault.runners.vault._validate_signature",
        autospec=True,
        return_value=None,
    ) as validate:
        yield validate


@pytest.mark.usefixtures("policies", "metadata")
@pytest.mark.parametrize(
    "config",
    [{}, {"issue:token:role_name": "test-role"}, {"issue:wrap": False}],
    indirect=True,
)
def test_generate_token(
    client_token,
    config,
    policies_default,
    token_serialized,
    wrapped_serialized,
    metadata_secret_default,
):
    """
    Ensure _generate_token calls the API as expected
    """
    wrap = config("issue:wrap")
    res_token, res_num_uses = vault._generate_token("test-minion", issue_params=None, wrap=wrap)
    endpoint = "auth/token/create"
    role_name = config("issue:token:role_name")
    payload = {}
    if config("issue:token:params:explicit_max_ttl"):
        payload["explicit_max_ttl"] = config("issue:token:params:explicit_max_ttl")
    if config("issue:token:params:num_uses"):
        payload["num_uses"] = config("issue:token:params:num_uses")
    payload["meta"] = metadata_secret_default
    payload["policies"] = policies_default
    if role_name:
        endpoint += f"/{role_name}"
    if config("issue:wrap"):
        assert res_token == wrapped_serialized
        client_token.post.assert_called_once_with(
            endpoint, payload=payload, wrap=config("issue:wrap")
        )
    else:
        res_token.pop("expire_time")
        assert res_token == token_serialized
    assert res_num_uses == 1


@pytest.mark.usefixtures("config", "policies")
@pytest.mark.parametrize("policies", [[]], indirect=True)
def test_generate_token_no_policies_denied():
    """
    Ensure generated tokens need at least one attached policy
    """
    with pytest.raises(salt.exceptions.SaltRunnerError, match=".*No policies matched minion.*"):
        vault._generate_token("test-minion", issue_params=None, wrap=False)


@pytest.mark.parametrize("ttl", [None, 1337])
@pytest.mark.parametrize("uses", [None, 1, 30])
@pytest.mark.parametrize("config", [{}, {"issue:type": "approle"}], indirect=True)
def test_generate_token_deprecated(ttl, uses, token_serialized, config, validate_signature, caplog):
    """
    Ensure the deprecated generate_token function returns data in the old format
    """
    issue_params = {}
    if ttl is not None:
        token_serialized["lease_duration"] = ttl
        issue_params["explicit_max_ttl"] = ttl
    if uses is not None:
        token_serialized["num_uses"] = uses
        issue_params["num_uses"] = uses
    expected = {
        "token": token_serialized["client_token"],
        "lease_duration": token_serialized["lease_duration"],
        "renewable": token_serialized["renewable"],
        "issued": token_serialized["creation_time"],
        "url": config("server:url"),
        "verify": config("server:verify"),
        "token_backend": config("cache:backend"),
        "namespace": config("server:namespace"),
        "uses": token_serialized["num_uses"],
    }
    with patch("saltext.vault.runners.vault._generate_token", autospec=True) as gen:
        gen.return_value = (token_serialized, token_serialized["num_uses"])
        res = vault.generate_token("test-minion", "sig", ttl=ttl, uses=uses)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with("test-minion", issue_params=issue_params or None, wrap=False)
        if config("issue:type") != "token":
            assert "Master is not configured to issue tokens" in caplog.text


@pytest.mark.parametrize("config", [{}, {"issue:wrap": False}], indirect=True)
@pytest.mark.parametrize("issue_params", [None, {"explicit_max_ttl": 120, "num_uses": 3}])
def test_generate_new_token(
    issue_params, config, validate_signature, token_serialized, wrapped_serialized
):
    """
    Ensure generate_new_token returns data as expected
    """
    if issue_params is not None:
        if issue_params.get("explicit_max_ttl") is not None:
            token_serialized["lease_duration"] = issue_params["explicit_max_ttl"]
        if issue_params.get("num_uses") is not None:
            token_serialized["num_uses"] = issue_params["num_uses"]
    expected = {"server": config("server"), "auth": {}}
    if config("issue:wrap"):
        expected.update(wrapped_serialized)
        expected.update({"misc_data": {"num_uses": token_serialized["num_uses"]}})
    else:
        expected["auth"] = token_serialized

    with patch("saltext.vault.runners.vault._generate_token", autospec=True) as gen:

        def res_or_wrap(*args, **kwargs):  # pylint: disable=unused-argument
            if kwargs.get("wrap"):
                return wrapped_serialized, token_serialized["num_uses"]
            return token_serialized, token_serialized["num_uses"]

        gen.side_effect = res_or_wrap
        res = vault.generate_new_token("test-minion", "sig", issue_params=issue_params)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with(
            "test-minion", issue_params=issue_params or None, wrap=config("issue:wrap")
        )


@pytest.mark.usefixtures("validate_signature", "config")
@pytest.mark.parametrize("config", [{"issue:type": "approle"}], indirect=True)
def test_generate_new_token_refuses_if_not_configured():
    """
    Ensure generate_new_token only issues tokens if configured to issue them
    """
    res = vault.generate_new_token("test-minion", "sig")
    assert "error" in res
    assert "Master does not issue tokens" in res["error"]


@pytest.mark.parametrize("config", [{}, {"issue:wrap": False}], indirect=True)
@pytest.mark.parametrize("issue_params", [None, {"explicit_max_ttl": 120, "num_uses": 3}])
def test_get_config_token(
    config, validate_signature, token_serialized, wrapped_serialized, issue_params
):
    """
    Ensure get_config returns data in the expected format when configured for token auth
    """
    expected = {
        "auth": {
            "method": "token",
            "token_lifecycle": {
                "minimum_ttl": 10,
                "renew_increment": None,
            },
        },
        "cache": config("cache"),
        "client": config("client"),
        "server": config("server"),
        "wrap_info_nested": [],
    }

    if issue_params is not None:
        if issue_params.get("explicit_max_ttl") is not None:
            token_serialized["lease_duration"] = issue_params["explicit_max_ttl"]
        if issue_params.get("num_uses") is not None:
            token_serialized["num_uses"] = issue_params["num_uses"]
    if config("issue:wrap"):
        expected["auth"].update({"token": wrapped_serialized})
        expected.update(
            {
                "wrap_info_nested": ["auth:token"],
                "misc_data": {"token:num_uses": token_serialized["num_uses"]},
            }
        )
    else:
        expected["auth"].update({"token": token_serialized})

    with patch("saltext.vault.runners.vault._generate_token", autospec=True) as gen:

        def res_or_wrap(*args, **kwargs):  # pylint: disable=unused-argument
            if kwargs.get("wrap"):
                return wrapped_serialized, token_serialized["num_uses"]
            return token_serialized, token_serialized["num_uses"]

        gen.side_effect = res_or_wrap
        res = vault.get_config("test-minion", "sig", issue_params=issue_params)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with(
            "test-minion", issue_params=issue_params or None, wrap=config("issue:wrap")
        )


@pytest.mark.parametrize(
    "config",
    [
        {"issue:type": "approle"},
        {
            "issue:type": "approle",
            "issue:wrap": False,
            "issue:approle:mount": "test-mount",
        },
        {"issue:type": "approle", "issue:approle:params:bind_secret_id": False},
    ],
    indirect=True,
)
@pytest.mark.parametrize(
    "issue_params",
    [
        None,
        {"token_explicit_max_ttl": 120, "token_num_uses": 3},
        {"secret_id_num_uses": 2, "secret_id_ttl": 120},
    ],
)
def test_get_config_approle(config, validate_signature, wrapped_serialized, issue_params):
    """
    Ensure get_config returns data in the expected format when configured for AppRole auth
    """
    expected = {
        "auth": {
            "approle_mount": config("issue:approle:mount"),
            "approle_name": "test-minion",
            "method": "approle",
            "secret_id": config("issue:approle:params:bind_secret_id"),
            "token_lifecycle": {
                "minimum_ttl": 10,
                "renew_increment": None,
            },
        },
        "cache": config("cache"),
        "client": config("client"),
        "server": config("server"),
        "wrap_info_nested": [],
    }

    if config("issue:wrap"):
        expected["auth"].update({"role_id": wrapped_serialized})
        expected.update({"wrap_info_nested": ["auth:role_id"]})
    else:
        expected["auth"].update({"role_id": "test-role-id"})

    with patch("saltext.vault.runners.vault._get_role_id", autospec=True) as gen:

        def res_or_wrap(*args, **kwargs):  # pylint: disable=unused-argument
            if kwargs.get("wrap"):
                return wrapped_serialized
            return "test-role-id"

        gen.side_effect = res_or_wrap
        res = vault.get_config("test-minion", "sig", issue_params=issue_params)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with(
            "test-minion", issue_params=issue_params or None, wrap=config("issue:wrap")
        )


@pytest.mark.parametrize(
    "config",
    [{"issue:type": "approle"}, {"issue:type": "approle", "issue:wrap": False}],
    indirect=True,
)
@pytest.mark.parametrize(
    "issue_params",
    [
        None,
        {"token_explicit_max_ttl": 120, "token_num_uses": 3},
        {"secret_id_num_uses": 2, "secret_id_ttl": 120},
    ],
)
def test_get_role_id(config, validate_signature, wrapped_serialized, issue_params):
    """
    Ensure get_role_id returns data in the expected format
    """
    expected = {"server": config("server"), "data": {}}
    if config("issue:wrap"):
        expected.update(wrapped_serialized)
    else:
        expected["data"].update({"role_id": "test-role-id"})
    with patch("saltext.vault.runners.vault._get_role_id", autospec=True) as gen:

        def res_or_wrap(*args, **kwargs):  # pylint: disable=unused-argument
            if kwargs.get("wrap"):
                return wrapped_serialized
            return "test-role-id"

        gen.side_effect = res_or_wrap
        res = vault.get_role_id("test-minion", "sig", issue_params=issue_params)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with(
            "test-minion", issue_params=issue_params or None, wrap=config("issue:wrap")
        )


@pytest.mark.usefixtures("validate_signature", "config")
@pytest.mark.parametrize("config", [{"issue:type": "token"}], indirect=True)
def test_get_role_id_refuses_if_not_configured():
    """
    Ensure get_role_id returns an error if not configured to issue AppRoles
    """
    res = vault.get_role_id("test-minion", "sig")
    assert "error" in res
    assert "Master does not issue AppRoles" in res["error"]


class TestGetRoleId:
    """
    Vault Testes Get RoleId
    """

    @pytest.fixture(autouse=True)
    def lookup_approle(self, approle_meta):
        with patch(
            "saltext.vault.runners.vault._lookup_approle_cached", autospec=True
        ) as lookup_approle:
            lookup_approle.return_value = approle_meta
            yield lookup_approle

    @pytest.fixture(autouse=True)
    def lookup_roleid(self, wrapped_serialized):
        role_id = MagicMock(return_value="test-role-id")
        role_id.serialize_for_minion.return_value = wrapped_serialized
        with patch("saltext.vault.runners.vault._lookup_role_id", autospec=True) as lookup_roleid:
            lookup_roleid.return_value = role_id
            yield lookup_roleid

    @pytest.fixture(autouse=True)
    def manage_approle(self):
        with patch("saltext.vault.runners.vault._manage_approle", autospec=True) as manage_approle:
            yield manage_approle

    @pytest.fixture(autouse=True)
    def manage_entity(self):
        with patch("saltext.vault.runners.vault._manage_entity", autospec=True) as manage_entity:
            yield manage_entity

    @pytest.fixture(autouse=True)
    def manage_entity_alias(self):
        with patch(
            "saltext.vault.runners.vault._manage_entity_alias", autospec=True
        ) as manage_entity_alias:
            yield manage_entity_alias

    @pytest.mark.parametrize(
        "config",
        [{"issue:type": "approle"}, {"issue:type": "approle", "issue:wrap": False}],
        indirect=True,
    )
    def test_get_role_id(
        self,
        config,
        lookup_approle,
        lookup_roleid,
        manage_approle,
        manage_entity,
        manage_entity_alias,
        wrapped_serialized,
    ):
        """
        Ensure _get_role_id returns data in the expected format and does not
        try to generate a new AppRole if it exists and is configured correctly
        """
        wrap = config("issue:wrap")
        res = vault._get_role_id("test-minion", issue_params=None, wrap=wrap)
        lookup_approle.assert_called_with("test-minion")
        lookup_roleid.assert_called_with("test-minion", wrap=wrap)
        manage_approle.assert_not_called()
        manage_entity.assert_not_called()
        manage_entity_alias.assert_not_called()

        if wrap:
            assert res == wrapped_serialized
            lookup_roleid.return_value.serialize_for_minion.assert_called_once()
        else:
            assert res() == "test-role-id"
            lookup_roleid.return_value.serialize_for_minion.assert_not_called()

    @pytest.mark.parametrize(
        "config",
        [
            {"issue:type": "approle"},
            {"issue:type": "approle", "issue:allow_minion_override_params": True},
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "issue_params", [None, {"token_explicit_max_ttl": 120, "token_num_uses": 3}]
    )
    def test_get_role_id_generate_new(
        self,
        config,
        lookup_approle,
        lookup_roleid,
        manage_approle,
        manage_entity,
        manage_entity_alias,
        wrapped_serialized,
        issue_params,
    ):
        """
        Ensure _get_role_id returns data in the expected format and does not
        try to generate a new AppRole if it exists and is configured correctly
        """
        lookup_approle.return_value = False
        wrap = config("issue:wrap")
        res = vault._get_role_id("test-minion", issue_params=issue_params, wrap=wrap)
        assert res == wrapped_serialized
        lookup_roleid.assert_called_with("test-minion", wrap=wrap)
        manage_approle.assert_called_once_with("test-minion", issue_params)
        manage_entity.assert_called_once_with("test-minion")
        manage_entity_alias.assert_called_once_with("test-minion")

    @pytest.mark.usefixtures("config")
    @pytest.mark.parametrize("config", [{"issue:type": "approle"}], indirect=True)
    def test_get_role_id_generate_new_errors_on_generation_failure(
        self, lookup_approle, lookup_roleid
    ):
        """
        Ensure _get_role_id returns an error if the AppRole generation failed
        """
        lookup_approle.return_value = False
        lookup_roleid.return_value = False
        with pytest.raises(
            salt.exceptions.SaltRunnerError,
            match="Failed to create AppRole for minion.*",
        ):
            vault._get_role_id("test-minion", issue_params=None, wrap=False)


@pytest.mark.parametrize(
    "config",
    [{"issue:type": "approle"}, {"issue:type": "approle", "issue:wrap": False}],
    indirect=True,
)
def test_generate_secret_id(
    config, validate_signature, wrapped_serialized, approle_meta, secret_id_serialized
):
    """
    Ensure generate_secret_id returns data in the expected format
    """
    expected = {
        "server": config("server"),
        "data": {},
        "misc_data": {"secret_id_num_uses": approle_meta["secret_id_num_uses"]},
    }
    if config("issue:wrap"):
        expected.update(wrapped_serialized)
    else:
        expected["data"].update(secret_id_serialized)
    with (
        patch("saltext.vault.runners.vault._get_secret_id", autospec=True) as gen,
        patch(
            "saltext.vault.runners.vault._approle_params_match",
            autospec=True,
            return_value=True,
        ) as matcher,
        patch(
            "saltext.vault.runners.vault._lookup_approle_cached", autospec=True
        ) as lookup_approle,
    ):

        def res_or_wrap(*args, **kwargs):  # pylint: disable=unused-argument
            if kwargs.get("wrap"):
                res = Mock(spec=vaultutil.VaultWrappedResponse)
                res.serialize_for_minion.return_value = wrapped_serialized
                return res
            secret_id = Mock(spec=vaultutil.VaultSecretId)
            secret_id.serialize_for_minion.return_value = secret_id_serialized
            return secret_id

        gen.side_effect = res_or_wrap
        lookup_approle.return_value = approle_meta
        res = vault.generate_secret_id("test-minion", "sig", issue_params=None)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with("test-minion", wrap=config("issue:wrap"))
        matcher.assert_called_once()


@pytest.mark.usefixtures("validate_signature", "config")
@pytest.mark.parametrize("config", [{"issue:type": "approle"}], indirect=True)
def test_generate_secret_id_nonexistent_approle():
    """
    Ensure generate_secret_id fails and prompts the minion to refresh cache if
    no associated AppRole could be found.
    """
    with patch(
        "saltext.vault.runners.vault._lookup_approle_cached", autospec=True
    ) as lookup_approle:
        lookup_approle.return_value = False
        res = vault.generate_secret_id("test-minion", "sig", issue_params=None)
        assert "error" in res
        assert "expire_cache" in res
        assert res["expire_cache"]


@pytest.mark.usefixtures("validate_signature", "config")
@pytest.mark.parametrize("config", [{"issue:type": "token"}], indirect=True)
def test_get_secret_id_refuses_if_not_configured():
    """
    Ensure get_secret_id returns an error if not configured to issue AppRoles
    """
    res = vault.generate_secret_id("test-minion", "sig")
    assert "error" in res
    assert "Master does not issue AppRoles" in res["error"]


@pytest.mark.parametrize("config", [{"issue:type": "approle"}], indirect=True)
def test_generate_secret_id_updates_params(
    config, validate_signature, wrapped_serialized, approle_meta
):
    """
    Ensure generate_secret_id returns data in the expected format
    """
    expected = {
        "server": config("server"),
        "data": {},
        "misc_data": {"secret_id_num_uses": approle_meta["secret_id_num_uses"]},
        "wrap_info": wrapped_serialized["wrap_info"],
    }
    with (
        patch("saltext.vault.runners.vault._get_secret_id", autospec=True) as gen,
        patch(
            "saltext.vault.runners.vault._approle_params_match",
            autospec=True,
            return_value=False,
        ) as matcher,
        patch("saltext.vault.runners.vault._manage_approle", autospec=True) as manage_approle,
        patch(
            "saltext.vault.runners.vault._lookup_approle_cached", autospec=True
        ) as lookup_approle,
    ):
        res = Mock(spec=vaultutil.VaultWrappedResponse)
        res.serialize_for_minion.return_value = wrapped_serialized
        gen.return_value = res
        lookup_approle.return_value = approle_meta
        res = vault.generate_secret_id("test-minion", "sig", issue_params=None)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with("test-minion", wrap=config("issue:wrap"))
        matcher.assert_called_once()
        manage_approle.assert_called_once()


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize("config", [{"issue:type": "token"}], indirect=True)
def test_list_approles_raises_exception_if_not_configured():
    """
    Ensure test_list_approles returns an error if not configured to issue AppRoles
    """
    with pytest.raises(salt.exceptions.SaltRunnerError, match="Master does not issue AppRoles.*"):
        vault.list_approles()


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize(
    "config,expected",
    [
        ({"policies:assign": ["no-tokens-to-replace"]}, ["no-tokens-to-replace"]),
        ({"policies:assign": ["single-dict:{minion}"]}, ["single-dict:test-minion"]),
        (
            {"policies:assign": ["should-not-cause-an-exception,but-result-empty:{foo}"]},
            [],
        ),
        (
            {"policies:assign": ["Case-Should-Be-Lowered:{grains[mixedcase]}"]},
            ["case-should-be-lowered:up-low-up"],
        ),
        (
            {"policies:assign": ["pillar-rendering:{pillar[role]}"]},
            ["pillar-rendering:test"],
        ),
    ],
    indirect=["config"],
)
def test_get_policies(expected, grains, pillar):
    """
    Ensure _get_policies works as intended.
    The expansion of lists is tested in the vault utility module unit tests.
    """
    with patch(
        "salt.utils.minions.get_minion_data",
        MagicMock(return_value=(None, grains, pillar)),
    ):
        with patch(
            "saltext.vault.utils.vault.helpers.expand_pattern_lists",
            Mock(side_effect=lambda x, *args, **kwargs: [x]),
        ):
            res = vault._get_policies("test-minion", refresh_pillar=False)
            assert res == expected


@pytest.mark.parametrize(
    "config",
    [
        {"policies:assign": ["salt_minion_{minion}"]},
        {"policies:assign": ["salt_grain_{grains[id]}"]},
        {"policies:assign": ["unset_{foo}"]},
        {"policies:assign": ["salt_pillar_{pillar[role]}"]},
    ],
    indirect=True,
)
def test_get_policies_does_not_render_pillar_unnecessarily(config, grains, pillar):
    """
    The pillar data should only be refreshed in case items are accessed.
    """
    with patch("salt.utils.minions.get_minion_data", autospec=True) as get_minion_data:
        get_minion_data.return_value = (None, grains, None)
        with patch(
            "saltext.vault.utils.vault.helpers.expand_pattern_lists",
            Mock(side_effect=lambda x, *args, **kwargs: [x]),
        ):
            with patch("salt.pillar.get_pillar", autospec=True) as get_pillar:
                get_pillar.return_value.compile_pillar.return_value = pillar
                vault._get_policies("test-minion", refresh_pillar=True)
                assert get_pillar.call_count == int("pillar" in config("policies:assign")[0])


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize(
    "config,expected",
    [
        ({"policies:assign": ["no-tokens-to-replace"]}, ["no-tokens-to-replace"]),
        ({"policies:assign": ["single-dict:{minion}"]}, ["single-dict:test-minion"]),
        ({"policies:assign": ["single-grain:{grains[os]}"]}, []),
    ],
    indirect=["config"],
)
def test_get_policies_for_nonexisting_minions(expected):
    """
    For non-existing minions, or the master-minion, grains will be None.
    """
    with patch("salt.utils.minions.get_minion_data", autospec=True) as get_minion_data:
        get_minion_data.return_value = (None, None, None)
        with patch(
            "saltext.vault.utils.vault.helpers.expand_pattern_lists",
            Mock(side_effect=lambda x, *args, **kwargs: [x]),
        ):
            res = vault._get_policies("test-minion", refresh_pillar=False)
            assert res == expected


@pytest.mark.parametrize(
    "metadata_patterns,expected",
    [
        (
            {"no-tokens-to-replace": "no-tokens-to-replace"},
            {"no-tokens-to-replace": "no-tokens-to-replace"},
        ),
        (
            {"single-dict:{minion}": "single-dict:{minion}"},
            {"single-dict:{minion}": "single-dict:test-minion"},
        ),
        (
            {"should-not-cause-an-exception,but-result-empty:{foo}": "empty:{foo}"},
            {"should-not-cause-an-exception,but-result-empty:{foo}": ""},
        ),
        (
            {"Case-Should-Not-Be-Lowered": "Case-Should-Not-Be-Lowered:{pillar[mixedcase]}"},
            {"Case-Should-Not-Be-Lowered": "Case-Should-Not-Be-Lowered:UP-low-UP"},
        ),
        (
            {"pillar-rendering:{pillar[role]}": "pillar-rendering:{pillar[role]}"},
            {"pillar-rendering:{pillar[role]}": "pillar-rendering:test"},
        ),
        (
            {"list-val": "list-val:{pillar[roles][0]}"},
            {"list-val": "list-val:foo"},
        ),
        (
            {"list-val-out-of-bounds": "list-val-out-of-bounds:{pillar[roles][10]}"},
            {"list-val-out-of-bounds": ""},
        ),
    ],
)
def test_get_metadata(metadata_patterns, expected, pillar):
    """
    Ensure _get_policies works as intended.
    The expansion of lists is tested in the vault utility module unit tests.
    """
    with patch("salt.utils.minions.get_minion_data", autospec=True) as get_minion_data:
        get_minion_data.return_value = (None, None, pillar)
        with patch(
            "saltext.vault.utils.vault.helpers.expand_pattern_lists",
            Mock(side_effect=lambda x, *args, **kwargs: [x]),
        ):
            res = vault._get_metadata("test-minion", metadata_patterns, refresh_pillar=False)
            assert res == expected


def test_get_metadata_list():
    """
    Test that lists are concatenated to an alphabetically sorted
    comma-separated list string since the API does not allow
    composite metadata values
    """
    with patch("salt.utils.minions.get_minion_data", autospec=True) as get_minion_data:
        get_minion_data.return_value = (None, None, None)
        with patch(
            "saltext.vault.utils.vault.helpers.expand_pattern_lists", autospec=True
        ) as expand:
            expand.return_value = ["salt_role_foo", "salt_role_bar"]
            res = vault._get_metadata(
                "test-minion",
                {"salt_role": "salt_role_{pillar[roles]}"},
                refresh_pillar=False,
            )
            assert res == {
                "salt_role": "salt_role_bar,salt_role_foo",
                "salt_role__0": "salt_role_bar",
                "salt_role__1": "salt_role_foo",
            }


def test_get_metadata_list_conflict():
    with patch("salt.utils.minions.get_minion_data", autospec=True) as get_minion_data:
        get_minion_data.return_value = (None, None, None)
        with patch(
            "saltext.vault.utils.vault.helpers.expand_pattern_lists", autospec=True
        ) as expand:
            expand.side_effect = lambda x, *args, **kwargs: ["foo", "bar"] if "pillar" in x else [x]
            res = vault._get_metadata(
                "test-minion",
                {"salt_role": "salt_role_{pillar[roles]}", "salt_role__1": "wat"},
                refresh_pillar=False,
            )
            assert res == {
                "salt_role": "bar,foo",
                "salt_role__1": "wat",
            }


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize(
    "config,issue_params,expected",
    [
        (
            {"issue:token:params": {"explicit_max_ttl": None, "num_uses": None}},
            None,
            {},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": 1337, "num_uses": None}},
            None,
            {"explicit_max_ttl": 1337},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": None, "num_uses": 3}},
            None,
            {"num_uses": 3},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": 1337, "num_uses": 3}},
            None,
            {"explicit_max_ttl": 1337, "num_uses": 3},
        ),
        (
            {
                "issue:token:params": {
                    "explicit_max_ttl": 1337,
                    "num_uses": 3,
                    "invalid": True,
                }
            },
            None,
            {"explicit_max_ttl": 1337, "num_uses": 3},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": None, "num_uses": None}},
            {"num_uses": 42, "explicit_max_ttl": 1338},
            {},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": 1337, "num_uses": None}},
            {"num_uses": 42, "explicit_max_ttl": 1338},
            {"explicit_max_ttl": 1337},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": None, "num_uses": 3}},
            {"num_uses": 42, "explicit_max_ttl": 1338},
            {"num_uses": 3},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": 1337, "num_uses": 3}},
            {"num_uses": 42, "explicit_max_ttl": 1338, "invalid": True},
            {"explicit_max_ttl": 1337, "num_uses": 3},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": None, "num_uses": None},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": None, "explicit_max_ttl": None},
            {},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": None, "num_uses": 3},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": 42, "explicit_max_ttl": None},
            {"num_uses": 42},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": 1337, "num_uses": None},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": None, "explicit_max_ttl": 1338},
            {"explicit_max_ttl": 1338},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": 1337, "num_uses": None},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": 42, "explicit_max_ttl": None},
            {"num_uses": 42, "explicit_max_ttl": 1337},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": None, "num_uses": 3},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": None, "explicit_max_ttl": 1338},
            {"num_uses": 3, "explicit_max_ttl": 1338},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": None, "num_uses": None},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": 42, "explicit_max_ttl": 1338},
            {"num_uses": 42, "explicit_max_ttl": 1338},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": 1337, "num_uses": 3},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": 42, "explicit_max_ttl": 1338, "invalid": True},
            {"num_uses": 42, "explicit_max_ttl": 1338},
        ),
        ({"issue:type": "approle", "issue:approle:params": {}}, None, {}),
        (
            {
                "issue:type": "approle",
                "issue:approle:params": {
                    "token_explicit_max_ttl": 1337,
                    "token_num_uses": 3,
                    "secret_id_num_uses": 3,
                    "secret_id_ttl": 60,
                },
            },
            None,
            {
                "token_explicit_max_ttl": 1337,
                "token_num_uses": 3,
                "secret_id_num_uses": 3,
                "secret_id_ttl": 60,
            },
        ),
        (
            {
                "issue:type": "approle",
                "issue:approle:params": {
                    "token_explicit_max_ttl": 1337,
                    "token_num_uses": 3,
                    "secret_id_num_uses": 3,
                    "secret_id_ttl": 60,
                },
            },
            {
                "token_explicit_max_ttl": 1338,
                "token_num_uses": 42,
                "secret_id_num_uses": 42,
                "secret_id_ttl": 1338,
            },
            {
                "token_explicit_max_ttl": 1337,
                "token_num_uses": 3,
                "secret_id_num_uses": 3,
                "secret_id_ttl": 60,
            },
        ),
        (
            {
                "issue:type": "approle",
                "issue:allow_minion_override_params": True,
                "issue:approle:params": {},
            },
            {
                "token_explicit_max_ttl": 1338,
                "token_num_uses": 42,
                "secret_id_num_uses": 42,
                "secret_id_ttl": 1338,
            },
            {
                "token_explicit_max_ttl": 1338,
                "token_num_uses": 42,
                "secret_id_num_uses": 42,
                "secret_id_ttl": 1338,
            },
        ),
        (
            {
                "issue:type": "approle",
                "issue:allow_minion_override_params": True,
                "issue:approle:params": {
                    "token_explicit_max_ttl": 1337,
                    "token_num_uses": 3,
                    "secret_id_num_uses": 3,
                    "secret_id_ttl": 60,
                },
            },
            {
                "token_explicit_max_ttl": 1338,
                "token_num_uses": 42,
                "secret_id_num_uses": 42,
                "secret_id_ttl": 1338,
            },
            {
                "token_explicit_max_ttl": 1338,
                "token_num_uses": 42,
                "secret_id_num_uses": 42,
                "secret_id_ttl": 1338,
            },
        ),
    ],
    indirect=["config"],
)
def test_parse_issue_params(issue_params, expected):
    """
    Ensure all known parameters can only be overridden if it was configured
    on the master. Also ensure the mapping to API requests is correct (for tokens).
    """
    res = vault._parse_issue_params(issue_params)
    assert res == expected


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize(
    "config,issue_params,expected",
    [
        (
            {"issue:type": "approle", "issue:approle:params": {}},
            {"bind_secret_id": False},
            False,
        ),
        (
            {"issue:type": "approle", "issue:approle:params": {}},
            {"bind_secret_id": True},
            False,
        ),
        (
            {"issue:type": "approle", "issue:approle:params": {"bind_secret_id": True}},
            {"bind_secret_id": False},
            True,
        ),
        (
            {
                "issue:type": "approle",
                "issue:approle:params": {"bind_secret_id": False},
            },
            {"bind_secret_id": True},
            False,
        ),
    ],
    indirect=["config"],
)
def test_parse_issue_params_does_not_allow_bind_secret_id_override(issue_params, expected):
    """
    Ensure bind_secret_id can only be set on the master.
    """
    res = vault._parse_issue_params(issue_params)
    assert res.get("bind_secret_id", False) == expected


@pytest.mark.usefixtures("config", "policies")
def test_manage_approle(approle_api, policies_default):
    """
    Ensure _manage_approle calls the API as expected.
    """
    vault._manage_approle("test-minion", None)
    approle_api.write_approle.assert_called_once_with(
        "test-minion",
        mount="salt-minions",
        explicit_max_ttl=9999999999,
        num_uses=1,
        token_policies=policies_default,
    )


@pytest.mark.usefixtures("config")
def test_delete_approle(approle_api):
    """
    Ensure _delete_approle calls the API as expected.
    """
    vault._delete_approle("test-minion")
    approle_api.delete_approle.assert_called_once_with("test-minion", mount="salt-minions")


@pytest.mark.usefixtures("config")
def test_lookup_approle(approle_api, approle_meta):
    """
    Ensure _lookup_approle calls the API as expected.
    """
    approle_api.read_approle.return_value = approle_meta
    res = vault._lookup_approle("test-minion")
    assert res == approle_meta
    approle_api.read_approle.assert_called_once_with("test-minion", mount="salt-minions")


@pytest.mark.usefixtures("config")
def test_lookup_approle_nonexistent(approle_api):
    """
    Ensure _lookup_approle catches VaultNotFoundErrors and returns False.
    """
    approle_api.read_approle.side_effect = vaultutil.VaultNotFoundError
    res = vault._lookup_approle("test-minion")
    assert res is False


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize("wrap", ["30s", False])
def test_lookup_role_id(approle_api, wrap):
    """
    Ensure _lookup_role_id calls the API as expected.
    """
    vault._lookup_role_id("test-minion", wrap=wrap)
    approle_api.read_role_id.assert_called_once_with("test-minion", mount="salt-minions", wrap=wrap)


@pytest.mark.usefixtures("config")
def test_lookup_role_id_nonexistent(approle_api):
    """
    Ensure _lookup_role_id catches VaultNotFoundErrors and returns False.
    """
    approle_api.read_role_id.side_effect = vaultutil.VaultNotFoundError
    res = vault._lookup_role_id("test-minion", wrap=False)
    assert res is False


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize("wrap", ["30s", False])
def test_get_secret_id(approle_api, wrap):
    """
    Ensure _get_secret_id calls the API as expected.
    """
    vault._get_secret_id("test-minion", wrap=wrap)
    approle_api.generate_secret_id.assert_called_once_with(
        "test-minion",
        metadata=ANY,
        wrap=wrap,
        mount="salt-minions",
    )


@pytest.mark.usefixtures("config")
def test_lookup_entity_by_alias(identity_api):
    """
    Ensure _lookup_entity_by_alias calls the API as expected.
    """
    with patch("saltext.vault.runners.vault._lookup_role_id", return_value="test-role-id"):
        vault._lookup_entity_by_alias("test-minion")
        identity_api.read_entity_by_alias.assert_called_once_with(
            alias="test-role-id", mount="salt-minions"
        )


@pytest.mark.usefixtures("config")
def test_lookup_entity_by_alias_failed(identity_api):
    """
    Ensure _lookup_entity_by_alias returns False if the lookup fails.
    """
    with patch("saltext.vault.runners.vault._lookup_role_id", return_value="test-role-id"):
        identity_api.read_entity_by_alias.side_effect = vaultutil.VaultNotFoundError
        res = vault._lookup_entity_by_alias("test-minion")
        assert res is False


@pytest.mark.usefixtures("config")
def test_fetch_entity_by_name(identity_api):
    """
    Ensure _fetch_entity_by_name calls the API as expected.
    """
    vault._fetch_entity_by_name("test-minion")
    identity_api.read_entity.assert_called_once_with(name="salt_minion_test-minion")


@pytest.mark.usefixtures("config")
def test_fetch_entity_by_name_failed(identity_api):
    """
    Ensure _fetch_entity_by_name returns False if the lookup fails.
    """
    identity_api.read_entity.side_effect = vaultutil.VaultNotFoundError
    res = vault._fetch_entity_by_name("test-minion")
    assert res is False


@pytest.mark.usefixtures("config", "metadata")
def test_manage_entity(identity_api, metadata_entity_default):
    """
    Ensure _manage_entity calls the API as expected.
    """
    vault._manage_entity("test-minion")
    identity_api.write_entity.assert_called_with(
        "salt_minion_test-minion", metadata=metadata_entity_default
    )


@pytest.mark.usefixtures("config")
def test_delete_entity(identity_api):
    """
    Ensure _delete_entity calls the API as expected.
    """
    vault._delete_entity("test-minion")
    identity_api.delete_entity.assert_called_with("salt_minion_test-minion")


@pytest.mark.usefixtures("config")
def test_manage_entity_alias(identity_api):
    """
    Ensure _manage_entity_alias calls the API as expected.
    """
    with patch("saltext.vault.runners.vault._lookup_role_id", return_value="test-role-id"):
        vault._manage_entity_alias("test-minion")
        identity_api.write_entity_alias.assert_called_with(
            "salt_minion_test-minion", alias_name="test-role-id", mount="salt-minions"
        )


@pytest.mark.usefixtures("config")
def test_manage_entity_alias_raises_errors(identity_api):
    """
    Ensure _manage_entity_alias raises exceptions.
    """
    identity_api.write_entity_alias.side_effect = vaultutil.VaultNotFoundError
    with patch("saltext.vault.runners.vault._lookup_role_id", return_value="test-role-id"):
        with pytest.raises(
            salt.exceptions.SaltRunnerError,
            match="Cannot create alias.* no entity found.",
        ):
            vault._manage_entity_alias("test-minion")


def test_revoke_token_by_token(client):
    """
    Ensure _revoke_token calls the API as expected.
    """
    vault._revoke_token(token="test-token")
    client.post.assert_called_once_with("auth/token/revoke", payload={"token": "test-token"})


def test_revoke_token_by_accessor(client):
    """
    Ensure _revoke_token calls the API as expected.
    """
    vault._revoke_token(accessor="test-accessor")
    client.post.assert_called_once_with(
        "auth/token/revoke-accessor", payload={"accessor": "test-accessor"}
    )


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize("cluster_id", (None, "test_cluster"))
@pytest.mark.parametrize("impersonated", (False, True))
def test_validate_signature_pki_dir(cluster_id, impersonated, master_opts):
    """
    Ensure we can validate minion signatures when running in
    master cluster mode.
    """
    master_opts["cluster_id"] = cluster_id
    base = Path(master_opts["pki_dir"])
    if not impersonated and cluster_id:
        base = base.parent / "cluster" / "pki"
        master_opts["cluster_pki_dir"] = str(base)
    if impersonated:
        expected = base / "master.pub"
    else:
        expected = base / "minions" / "test-minion"
    with patch("salt.crypt.verify_signature", autospec=True) as verify:
        vault._validate_signature("test-minion", "", impersonated)
    assert verify.call_args.args[0] == str(expected)


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize("config", ({"issue:block_minion_requests": True},), indirect=True)
@pytest.mark.parametrize("impersonated", (False, True))
def test_issue_non_impersonated_only(impersonated):
    if not impersonated:
        ctx = pytest.raises(salt.exceptions.AuthenticationError, match="have been disabled")
    else:
        ctx = contextlib.nullcontext()
    with ctx, patch("salt.crypt.verify_signature", autospec=True) as verify:
        vault._validate_signature("test-minion", "", impersonated)
    assert bool(verify.call_count) is impersonated



import pytest

from tests.support.vault import vault_delete
from tests.support.vault import vault_disable_secret_engine
from tests.support.vault import vault_enable_secret_engine
from tests.support.vault import vault_list
from tests.support.vault import vault_read
from tests.support.vault import vault_write

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        },
    }


@pytest.fixture
def userrole():
    return {
        "key_type": "ca",
        "allowed_users": "foo,bar,baz",
        "allowed_extensions": "*",
        "allow_user_certificates": True,
        "ttl": 3600,
        "max_ttl": 86400,
    }


@pytest.fixture
def hostrole():
    return {
        "key_type": "ca",
        "allowed_domains": "*",
        "allow_host_certificates": True,
        "ttl": 3600,
        "max_ttl": 86400,
    }


@pytest.fixture
def iprole():
    return {
        "key_type": "otp",
        "default_user": "foobar",
        "cidr_list": "0.0.0.0/1",
        "exclude_cidr_list": "128.0.0.0/1",
        "port": 9876,
    }


@pytest.fixture
def ec_priv():
    return """
-----BEGIN OPENSSH PRIVATE KEY-----
b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAAAaAAAABNlY2RzYS1zaGEy
LW5pc3RwMjU2AAAACG5pc3RwMjU2AAAAQQSO2hM3nJP6fxgzyXIEEKhHeqOqXccIvV8EZLfqCrcX
NGR7Be7yYdPNx+bcNFx5fyLrxKin+f/4pV4/q+mMoqVxAAAAoHwE+2V8BPtlAAAAE2VjZHNhLXNo
YTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBI7aEzeck/p/GDPJcgQQqEd6o6pdxwi9XwRkt+oK
txc0ZHsF7vJh083H5tw0XHl/IuvEqKf5//ilXj+r6YyipXEAAAAgc5sbtq4PEIHmkqJzEbNaO1sB
2PSfCjWqlGSC7ODBqy4AAAAAAQIDBAUGBwg=
-----END OPENSSH PRIVATE KEY-----
    """.strip()


@pytest.fixture
def ec_priv_file(ec_priv, tmp_path):
    path = tmp_path / "ec"
    path.write_text(ec_priv)
    return str(path)


@pytest.fixture
def ec_pub():
    return "ecdsa-sha2-nistp256 AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBI7aEzeck/p/GDPJcgQQqEd6o6pdxwi9XwRkt+oKtxc0ZHsF7vJh083H5tw0XHl/IuvEqKf5//ilXj+r6YyipXE="


@pytest.fixture
def ec_pub_file(ec_pub, tmp_path):
    path = tmp_path / "ec.pub"
    path.write_text(ec_pub)
    return str(path)


@pytest.fixture(scope="module", autouse=True)
def ssh_engine(container):  # pylint: disable=unused-argument
    assert vault_enable_secret_engine("ssh")
    yield
    assert vault_disable_secret_engine("ssh")


@pytest.fixture
def roles_clean():
    try:
        yield
    finally:
        for role_name in ("userrole", "hostrole", "iprole"):
            if role_name in vault_list("ssh/roles"):
                vault_delete(f"ssh/roles/{role_name}")
                assert role_name not in vault_list("ssh/roles")


@pytest.fixture(params=(("userrole", "hostrole", "iprole"),))
def roles_setup(request, roles_clean):  # pylint: disable=unused-argument
    for role_name in request.param:
        role_args = request.getfixturevalue(role_name)
        vault_write(f"ssh/roles/{role_name}", **role_args)
        assert role_name in vault_list("ssh/roles")
    yield


@pytest.fixture
def vault_ssh(states):
    try:
        yield states.vault_ssh
    finally:
        pass


@pytest.fixture
def _temp_role():
    name = "testrole"
    try:
        yield name
    finally:
        vault_delete(f"ssh/roles/{name}")


@pytest.fixture
def _temp_ca():
    try:
        yield
    finally:
        vault_delete("ssh/config/ca")


@pytest.fixture
def ca_setup(ec_priv, ec_pub):
    vault_write("ssh/config/ca", private_key=ec_priv, public_key=ec_pub)
    assert vault_read("ssh/config/ca", default=False)
    try:
        yield
    finally:
        vault_delete("ssh/config/ca")


@pytest.fixture(params=(False, True))
def testmode(request):
    return request.param


@pytest.mark.usefixtures("_temp_ca")
def test_ca_present(vault_ssh, testmode):
    ret = vault_ssh.ca_present("foobar", test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("would have" in ret.comment) is testmode
    assert "created" in ret.changes
    assert bool(vault_read("ssh/config/ca", default=False)) is not testmode


@pytest.mark.usefixtures("ca_setup")
def test_ca_present_already_present(vault_ssh, testmode):
    ret = vault_ssh.ca_present("foobar", test=testmode)
    assert ret.result is True
    assert "already been initialized" in ret.comment
    assert not ret.changes


@pytest.mark.usefixtures("ca_setup")
def test_ca_absent(vault_ssh, testmode):
    ret = vault_ssh.ca_absent("foobar", test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("would have" in ret.comment) is testmode
    assert "destroyed" in ret.changes
    assert bool(vault_read("ssh/config/ca", default=False)) is testmode


def test_ca_absent_already_absent(vault_ssh, testmode):
    ret = vault_ssh.ca_absent("foobar", test=testmode)
    assert ret.result is True
    assert "There is no CA" in ret.comment
    assert not ret.changes


@pytest.mark.usefixtures("roles_clean")
def test_role_present_ca(vault_ssh, userrole, testmode):
    key_type = userrole.pop("key_type")
    ret = vault_ssh.role_present_ca("userrole", **userrole, test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("would have" in ret.comment) is testmode
    assert "created" in ret.changes
    assert bool(vault_read("ssh/roles/userrole", default=False)) is not testmode
    if not testmode:
        data = vault_read("ssh/roles/userrole")["data"]
        assert data["key_type"] == key_type
        for param, val in userrole.items():
            assert data[param] == val


@pytest.mark.usefixtures("roles_clean")
def test_role_present_otp(vault_ssh, iprole, testmode):
    key_type = iprole.pop("key_type")
    ret = vault_ssh.role_present_otp("iprole", **iprole, test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("would have" in ret.comment) is testmode
    assert "created" in ret.changes
    assert bool(vault_read("ssh/roles/iprole", default=False)) is not testmode
    if not testmode:
        data = vault_read("ssh/roles/iprole")["data"]
        assert data["key_type"] == key_type
        for param, val in iprole.items():
            assert data[param] == val


@pytest.mark.usefixtures("roles_setup")
def test_role_present_ca_already_present(vault_ssh, userrole, testmode):
    userrole.pop("key_type")
    ret = vault_ssh.role_present_ca("userrole", **userrole, test=testmode)
    assert ret.result is True
    assert "as specified" in ret.comment
    assert not ret.changes


@pytest.mark.usefixtures("roles_setup")
def test_role_present_otp_already_present(vault_ssh, iprole, testmode):
    iprole.pop("key_type")
    ret = vault_ssh.role_present_otp("iprole", **iprole, test=testmode)
    assert ret.result is True
    assert "as specified" in ret.comment
    assert not ret.changes


@pytest.mark.usefixtures("roles_setup")
def test_role_present_ca_changes(vault_ssh, userrole, testmode):
    change = {
        "allow_user_key_ids": True,
        "ttl": "2h",
        "allowed_user_key_lengths": {"rsa": [2048, 3072], "ec": 256},
    }
    params = userrole.copy()
    params.update(change)
    params.pop("key_type")
    ret = vault_ssh.role_present_ca("userrole", **params, test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("would have" in ret.comment) is testmode
    assert set(ret.changes) == set(change)
    assert ret["changes"]["ttl"] == {"old": userrole.get("ttl", 0), "new": 7200}
    assert ret["changes"]["allow_user_key_ids"] == {
        "old": userrole.get("allow_user_key_ids", False),
        "new": change["allow_user_key_ids"],
    }
    assert ret["changes"]["allowed_user_key_lengths"] == {
        "added": ["ec", "rsa"],
        "changed": [],
        "removed": [],
    }

    new = vault_read("ssh/roles/userrole")["data"]
    assert (new["ttl"] != 7200) is testmode
    assert new["allow_user_key_ids"] is not testmode


@pytest.mark.usefixtures("roles_setup")
def test_role_present_otp_changes(vault_ssh, iprole, testmode):
    change = {
        "default_user": "barbaz",
        "cidr_list": ["0.0.0.0/1", "128.0.0.0/1"],
        "exclude_cidr_list": None,
    }
    iprole.update(change)
    iprole.pop("key_type")
    ret = vault_ssh.role_present_otp("iprole", **iprole, test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("would have" in ret.comment) is testmode
    assert set(ret.changes) == set(change)
    assert ret.changes["default_user"] == {"old": "foobar", "new": "barbaz"}
    assert ret.changes["cidr_list"] == {"added": ["128.0.0.0/1"], "removed": []}
    assert ret.changes["exclude_cidr_list"] == {"added": [], "removed": ["128.0.0.0/1"]}

    new = vault_read("ssh/roles/iprole")["data"]
    assert (new["default_user"] != "barbaz") is testmode
    assert (new["cidr_list"] != "0.0.0.0/1,128.0.0.0/1") is testmode
    assert bool(new["exclude_cidr_list"]) is testmode


@pytest.mark.usefixtures("roles_setup")
def test_role_present_key_type_change(vault_ssh, iprole, testmode):
    iprole.pop("key_type")
    ret = vault_ssh.role_present_otp("userrole", **iprole, test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("would have" in ret.comment) is testmode
    assert set(ret.changes) == {"key_type"}
    assert ret.changes["key_type"] == {"old": "ca", "new": "otp"}
    new = vault_read("ssh/roles/userrole")["data"]
    assert (new["key_type"] == "ca") is testmode


@pytest.mark.usefixtures("roles_setup")
def test_role_absent(vault_ssh, testmode):
    ret = vault_ssh.role_absent("userrole", test=testmode)
    assert ret.result is (None if testmode else True)
    assert ("would have" in ret.comment) is testmode
    assert "deleted" in ret.changes
    assert bool(vault_read("ssh/roles/userrole", default=False)) is testmode


def test_role_absent_already_absent(vault_ssh, testmode):
    ret = vault_ssh.role_absent("foobar", test=testmode)
    assert ret.result is True
    assert "already absent" in ret.comment
    assert not ret.changes



"""
Manage the Vault (or OpenBao) PKI secret engine, request X.509 certificates.

.. versionadded:: 1.1.0

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.
"""

import logging

from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltInvocationError

from saltext.vault.utils import vault
from saltext.vault.utils.vault.pki import dec2hex

log = logging.getLogger(__name__)

__virtualname__ = "vault_pki"


def __virtual__():
    return __virtualname__


VALID_CSR_ARGS = (
    "C",
    "ST",
    "L",
    "STREET",
    "O",
    "OU",
    "CN",
    "MAIL",
    "SN",
    "GN",
    "UID",
    "authorityKeyIdentifier",
    "basicConstraints",
    "certificatePolicies",
    "extendedKeyUsage",
    "inhibitAnyPolicy",
    "keyUsage",
    "nameConstraints",
    "noCheck",
    "policyConstraints",
    "subjectKeyIdentifier",
    "tlsfeature",
)


def list_roles(mount="pki"):
    """
    List configured PKI roles.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#list-roles>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.list_roles

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """
    endpoint = f"{mount}/roles"
    try:
        return vault.query("LIST", endpoint, __opts__, __context__)["data"]["keys"]
    except vault.VaultNotFoundError:
        return []
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def read_role(name, mount="pki"):
    """
    Get configuration of specific PKI role.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#read-role>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.read_role

    name
        The name of the role.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """

    endpoint = f"{mount}/roles/{name}"
    try:
        res = vault.query("GET", endpoint, __opts__, __context__)
        return res["data"]
    except vault.VaultNotFoundError:
        return None
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def write_role(
    name,
    mount="pki",
    issuer_ref=None,
    ttl=None,
    max_ttl=None,
    allow_localhost=None,
    allowed_domains=None,
    server_flag=None,
    client_flag=None,
    key_usage=None,
    no_store=None,
    require_cn=None,
    **kwargs,
):
    """
    Create or update PKI role.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#create-update-role>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.write_role myrole

    name
        The name of the role.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    issuer_ref
        Name or id of the issuer which will be used with this role. If not set, default issuer will be used.

    ttl
        Specifies the Time To Live value to be used for the validity period of the requested certificate,
        provided as a string duration with time suffix. Hour is the largest suffix.
        The value specified is strictly used for future validity.
        If not set, uses the system default value or the value of ``max_ttl``, whichever is shorter.

    max_ttl
        Specifies the maximum Time To Live provided as a string duration with time suffix.
        Hour is the largest suffix. If not set, defaults to the system maximum lease TTL.

    allow_localhost
        Specifies if clients can request certificates for ``localhost`` as one of the requested common names.

    allowed_domains
        Specifies the domains this role is allowed to issue certificates for.
        This is used with the ``allow_bare_domains``, ``allow_subdomains``, and ``allow_glob_domains`` options to
        determine the type of matching between these domains and the values of common name, DNS-typed SAN entries, and Email-typed SAN entries.
        When ``allow_any_name`` is used, this attribute has no effect.

    server_flag
        Specifies if certificates are flagged for server authentication use.
        See `RFC 5280 Section 4.2.1.12 <https://datatracker.ietf.org/doc/html/rfc5280#section-4.2.1.12>`__
        for information about the Extended Key Usage field.
        If not set, defaults to true.

    client_flag
        Specifies if certificates are flagged for client authentication use.
        See `RFC 5280 Section 4.2.1.12 <https://datatracker.ietf.org/doc/html/rfc5280#section-4.2.1.12>`__
        for information about the Extended Key Usage field.
        If not set, defaults to true.

    key_usage
        Specifies the allowed key usage constraint on issued certificates.
        If unset, defaults to ``["DigitalSignature", "KeyAgreement", "KeyEncipherment"]``

    no_store
        If set, certificates issued/signed against this role will not be stored in the storage backend.

    require_cn
        If set to false, makes the common_name field optional while generating a certificate. Defaults to true.

    kwargs:
        Any other params which can be understand by Vault API.

    """

    endpoint = f"{mount}/roles/{name}"
    method = "POST"

    if read_role(name, mount=mount) is not None:
        method = "PATCH"

    payload = {k: v for k, v in kwargs.items() if not k.startswith("_")}

    if issuer_ref is not None:
        payload["issuer_ref"] = issuer_ref
    if ttl is not None:
        payload["ttl"] = ttl
    if max_ttl is not None:
        payload["max_ttl"] = max_ttl
    if allow_localhost is not None:
        payload["allow_localhost"] = allow_localhost
    if allowed_domains is not None:
        if not isinstance(allowed_domains, list):
            allowed_domains = [allowed_domains]
        payload["allowed_domains"] = allowed_domains
    if server_flag is not None:
        payload["server_flag"] = server_flag
    if client_flag is not None:
        payload["client_flag"] = client_flag
    if key_usage is not None:
        if not isinstance(key_usage, list):
            key_usage = [key_usage]
        payload["key_usage"] = key_usage
    if no_store is not None:
        payload["no_store"] = no_store
    if require_cn is not None:
        payload["require_cn"] = require_cn

    try:
        vault.query(method, endpoint, __opts__, __context__, payload=payload, safe_to_retry=True)
        return True
    except vault.VaultUnsupportedOperationError as err:
        raise CommandExecutionError(
            f"Vault version too old. Please upgrade to v1.11.0+: {err}"
        ) from err
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def delete_role(name, mount="pki"):
    """
    Delete PKI role from Vault.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#delete-role>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.delete_role myrole

    name
        The name of the role.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """

    endpoint = f"{mount}/roles/{name}"

    try:
        vault.query("DELETE", endpoint, __opts__, __context__)
        return True
    except vault.VaultNotFoundError:
        return False
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def list_issuers(mount="pki"):
    """
    List issuers information
    Returns ``{ "<issuer_id>" : { "is_default": False, "issuer_name": "...", "key_id": "...", "serial_number": "...."}}``


    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#list-issuers>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.list_issuers

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """
    endpoint = f"{mount}/issuers"

    try:
        return vault.query("LIST", endpoint, __opts__, __context__, is_unauthd=True)["data"][
            "key_info"
        ]
    except vault.VaultNotFoundError:
        return []
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def read_issuer(ref="default", mount="pki"):
    """
    Read an issuer's information.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#read-issuer-certificate>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.read_issuer

    ref
        Reference of the issuer. Can be issuer id, issuer name or literal ``default``
        which means default issuer. Defaults to ``default``.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    """
    endpoint = f"{mount}/issuer/{ref}"

    try:
        return vault.query("GET", endpoint, __opts__, __context__, is_unauthd=True)["data"]
    except vault.VaultNotFoundError:
        return None
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def update_issuer(
    ref="default",
    mount="pki",
    manual_chain=None,
    usage=None,
    aia_urls=None,
    crl_endpoints=None,
    ocsp_servers=None,
):
    """
    Update issuer's information.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#update-issuer>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.update_issuer ref usage=["crl-signing"]

    ref
        Reference of the issuer. Can be issuer id, issuer name or literal ``default``
        which means default issuer. Defaults to ``default``.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    manual_chain
        Chain of issuer references to build this issuer's computed CAChain field from, when non-empty.

    usage
        Allowed usages for this issuer. Valid options are:

        * ``read-only`` - to allow this issuer to be read; implict; always allowed;
        * ``issuing-certificates`` - to allow this issuer to be used for issuing other certificates;
        * ``crl-signing`` -  to allow this issuer to be used for signing CRLs.
          This is separate from the CRLSign KeyUsage on the x509 certificate, but this usage cannot be set
          unless that KeyUsage is allowed on the x509 certificate;
        * ``ocsp-signing`` -  to allow this issuer to be used for signing OCSP responses.


    aia_urls
        Specifies the URL values for the Issuing Certificate field as an array.

    crl_endpoints
        Specifies the URL values for the CRL Distribution Points field as an array.

    ocsp_servers
        Specifies the URL values for the OCSP Servers field as an array.

    """
    endpoint = f"{mount}/issuer/{ref}"

    payload = {}

    if manual_chain is not None:
        payload["manual_chain"] = manual_chain

    if usage:
        payload["usage"] = usage

    if aia_urls is not None:
        payload["issuing_certificates"] = aia_urls

    if crl_endpoints is not None:
        payload["crl_distribution_points"] = crl_endpoints

    if ocsp_servers is not None:
        payload["ocsp_servers"] = ocsp_servers

    try:
        vault.query(
            "PATCH",
            endpoint,
            __opts__,
            __context__,
            payload=payload,
        )
        return True
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def read_issuer_certificate(name="default", mount="pki", include_chain=False):
    """
    Read an issuer's certificate.
    Returns certificate(s) in PEM format

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#read-issuer-certificate>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.read_issuer_certificate

    name
        Name of the issuer. Can be issuer id, issuer name or literal ``default``
        which means default issuer. Defaults to ``default``.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    include_chain
        If set to true will append the CA chain to the certificate (in case of intermediate issuer)
    """
    cert_data = read_issuer(name, mount)

    if include_chain:
        return "".join(cert_data["ca_chain"])
    return cert_data["certificate"]


def get_default_issuer(mount="pki"):
    """
    Return the issuer ID of the default issuer.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#list-issuers>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.get_default_issuer

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """
    all_issuers = list_issuers(mount)

    for k, v in all_issuers.items():
        if v["is_default"]:
            return k
    # In case there is no default issuer
    return None


def set_default_issuer(name, mount="pki"):
    """
    Set the default issuer.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#set-issuers-configuration>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.set_default_issuer myca

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """
    endpoint = f"{mount}/config/issuers"
    payload = {"default": name}
    try:
        vault.query("POST", endpoint, __opts__, __context__, payload=payload, safe_to_retry=True)
        return True
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def generate_root(
    common_name,
    mount="pki",
    type="internal",  # pylint: disable=redefined-builtin
    issuer_name=None,
    key_name=None,
    ttl=None,
    key_type="rsa",
    key_bits=0,
    max_path_length=-1,
    **kwargs,
):
    """
    Generate a new root issuer.
    Returns ``{ "certificate" : "-----BEGIN CERTIFICATE...", "issuer_id": "...", "key_id": "...", }``
    If type is ``exported`` it will also return the private key.


    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#generate-root>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.generate_root my-root

    common_name
        The common name to be used for the CA

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    type
        Specifies the type of the root to create. If ``exported``, the private key will be returned in the response;
        if ``internal``, the private key will not be returned and cannot be retrieved later. Defaults to ``internal``.

    issuer_name
        Provides a name to the specified issuer. The name must be unique across all issuers and not be the reserved value ``default``.

    key_name
        When a new key is created with this request, optionally specifies the name for this. The global ref ``default`` may not be used as a name.

    ttl
        Specifies the requested Time To Live (after which the certificate will be expired). This cannot be larger than the engine's max (or, if not set, the system max).

    key_type
        Specifies the desired key type; must be ``rsa``, ``ed25519`` or ``ec``. Defaults to ``rsa``.

    key_bits
        Specifies the number of bits to use for the generated keys.
        Allowed values are 0 (universal default);
        with ``key_type=rsa``, allowed values are: 2048 (default), 3072, 4096 or 8192;
        with ``key_type=ec``, allowed values are: 224, 256 (default), 384, or 521;
        ignored with ``key_type=ed25519``.

    max_path_length
        Specifies the maximum path length to encode in the generated certificate. ``-1`` means no limit,
        unless the signing certificate has a maximum path length set, in which case the path length is set to one
        less than that of the signing certificate. A limit of 0 means a literal path length of zero.
    """

    if issuer_name == "default":
        raise SaltInvocationError("issuer_name cannot be `default`. This is a reserved word.")

    if key_name == "default":
        raise SaltInvocationError("key_name cannot be `default`. This is a reserved word.")

    endpoint = f"{mount}/root/generate/{type}"

    payload = {k: v for k, v in kwargs.items() if not k.startswith("_")}

    payload["common_name"] = common_name
    payload["key_type"] = key_type

    if issuer_name is not None:
        payload["issuer_name"] = issuer_name
    if key_name is not None:
        payload["key_name"] = key_name
    if ttl is not None:
        payload["ttl"] = ttl

    if key_bits > 0:
        payload["key_bits"] = key_bits

    if max_path_length > -1:
        payload["max_path_length"] = max_path_length

    try:
        resp = vault.query("POST", endpoint, __opts__, __context__, payload=payload)["data"]
        ret = {
            "certificate": resp["certificate"],
            "issuer_id": resp["issuer_id"],
            "key_id": resp["key_id"],
        }

        if type == "exported":
            ret["private_key"] = resp["private_key"]

        return ret
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def delete_key(ref, mount="pki"):
    """
    Delete private key from Vault.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#delete-key>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.delete_key ref

    ref
        Ref of the key. Could be name or key_id.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """

    endpoint = f"{mount}/key/{ref}"
    try:
        vault.query("DELETE", endpoint, __opts__, __context__)
        return True
    except vault.VaultNotFoundError:
        return False
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def delete_issuer(ref, mount="pki", include_key=False):
    """
    Delete issuer from Vault.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#delete-issuer>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.delete_issuer ref

    ref
        Ref of the issuer. Could be name or issuer_id.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    include_key
        If set to true will also delete the private key if imported.
        Defaults to false, so private key will be preserved.
    """

    endpoint = f"{mount}/issuer/{ref}"
    key_id = None

    if include_key:
        issuer_info = read_issuer(ref, mount=mount)
        if issuer_info:
            key_id = issuer_info["key_id"]

    try:
        vault.query("DELETE", endpoint, __opts__, __context__)
        if key_id:
            delete_key(key_id, mount=mount)
        return True
    except vault.VaultNotFoundError:
        return False
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def read_issuer_crl(ref="default", mount="pki", delta=False):
    """
    Get issuer CRL.

    .. note::
        If CA cannot sign CRLs will return None.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#read-issuer-crl>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.read_issuer_crl ref

    ref
        Ref of the issuer. Could be name or issuer_id. Defaults to default issuer.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    delta
        If set to true, will return delta CRL instead of complete one.
    """
    # Check if issuer can sign CRLs at all. If not,
    # there is no point to check for CRL as this will throw error
    issuer = None
    try:
        issuer = vault.query(
            "GET", f"{mount}/issuer/{ref}", __opts__, __context__, is_unauthd=False
        )["data"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err

    if issuer is None:
        return None

    if "crl-signing" not in issuer["usage"].split(","):
        return None

    endpoint = f"{mount}/issuer/{ref}/crl"
    if delta:
        endpoint = endpoint + "/delta"

    try:
        return vault.query("GET", endpoint, __opts__, __context__, is_unauthd=True)["data"]["crl"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def list_revoked_certificates(mount="pki"):
    """
    List revoked certificates serial numbers

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#list-revoked-certificates>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.list_revoked_certificates

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """
    endpoint = f"{mount}/certs/revoked"

    try:
        return vault.query("LIST", endpoint, __opts__, __context__)["data"]["keys"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def list_certificates(mount="pki"):
    """
    List issued certificates serial numbers

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#list-certificates>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.list_certificates

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """
    endpoint = f"{mount}/certs"

    try:
        return vault.query("LIST", endpoint, __opts__, __context__)["data"]["keys"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def read_certificate(serial, mount="pki"):
    """
    Read issued certificate.
    Returns certificate in PEM format

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#read-certificate>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.read_certificate 7e:85:c5:d1:85:94:9a:46:08:b5:1b:9c:22:cb:35:e5:ea:f3:56:3f

    serial
        Specifies the serial of the key to read. Valid values are:

        * ``<serial>`` for the certificate with the given serial number, in hyphen-separated or colon-separated hexadecimal.
        * ``ca`` for the default issuer's CA certificate
        * ``crl`` for the default issuer's CRL
        * ``ca_chain`` for the default issuer's CA trust chain.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """
    endpoint = f"{mount}/cert/{serial}"

    try:
        return vault.query("GET", endpoint, __opts__, __context__)["data"]["certificate"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def issue_certificate(
    role_name,
    common_name,
    mount="pki",
    issuer_ref=None,
    alt_names=None,
    ttl=None,
    format="pem",  # pylint: disable=redefined-builtin
    exclude_cn_from_sans=False,
    **kwargs,
):
    """
    Generate and issue a new certificate with private key.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#generate-certificate-and-key>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.issue_certificate myrole common_name="www.example.com"

    role_name
        Name of the role to be used for issuing the certificate.

    common_name
        Common name to be set for the certificate.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    issuer_ref
        Override role's issuer. Can be issuer_name or issuer_id.

    alt_names
        Any alternative names to be added to the certificate. Can be specified either as dict (``{ "<type>": "<value"}``)
        or list of SANs (``["<type>:<value>"]``).

    ttl
        Specifies the requested Time To Live (after which the certificate will be expired).
        This cannot be larger than the engine's max (or, if not set, the system max).

    format
        Can be either ``pem`` or ``der``. Defaults to ``pem``.

    exclude_cn_from_sans
        If set to true, Common name will not be part of the SANs.

    kwargs
        Any additional parameter accepted by Vault API.
    """
    endpoint = f"{mount}/issue/{role_name}"
    if issuer_ref is not None:
        endpoint = f"{mount}/issuer/{issuer_ref}/issue/{role_name}"

    payload = {k: v for k, v in kwargs.items() if not k.startswith("_")}
    payload["common_name"] = common_name

    if ttl is not None:
        payload["ttl"] = ttl

    payload["format"] = format
    payload["exclude_cn_from_sans"] = exclude_cn_from_sans

    if alt_names is not None:
        dns_sans, ip_sans, uri_sans, other_sans = _split_sans(alt_names)
        payload["alt_names"] = ",".join(dns_sans)
        payload["ip_sans"] = ",".join(ip_sans)
        payload["uri_sans"] = ",".join(uri_sans)
        payload["other_sans"] = ",".join(other_sans)

    try:
        return vault.query("POST", endpoint, __opts__, __context__, payload=payload)["data"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def sign_certificate(
    role_name,
    common_name,
    mount="pki",
    csr=None,
    private_key=None,
    private_key_passphrase=None,
    digest="sha256",
    issuer_ref=None,
    alt_names=None,
    ttl=None,
    sign_verbatim=False,
    encoding="pem",
    exclude_cn_from_sans=False,
    **kwargs,
):
    """
    Issue a new certificate from existing private key or CSR.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#sign-certificate>`__.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#sign-verbatim>`__

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.issue_certificate myrole common_name="www.example.com"

    role_name
        Name of the role to be used for issuing the certificate.

    common_name
        Common name to be set for the certificate.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.

    csr
        Pass the CSR which will be used for issuing the certificate. Either ``csr`` or ``private_key`` parameter can be set, not both.

    private_key
        The private key for which certificate should be issued. Can be text or path.
        Either ``csr`` or ``private_key`` parameter can be set, not both.

        .. note::
            This parameter requires the :py:mod:`x509_v2 execution module <salt.modules.x509_v2>` to be available.

    private_key_passphrase
        The passphrase for the ``private_key`` if encrypted. Not used in case of ``csr``.

    digest
        Digest to be used for generating the CSR. Not used in case of ``private_key``. Defaults to ``sha256``

    issuer_ref
        Override role's issuer. Can be issuer_name or issuer_id.

    alt_names
        Any alternative names to be added to the certificate. Can be specified either as dict (``{ "<type>": "<value"}``)
        or list of SANs (``["<type>:<value>"]``).

    ttl
        Specifies the requested Time To Live (after which the certificate will be expired).
        This cannot be larger than the engine's max (or, if not set, the system max).

    sign_verbatim
        If set to true, the resulting certificate follows the CSR exactly.
        Otherwise, only ``CN`` can be set for the subject, any other subject parameter (like ``O``) is ignored.

        .. warning::
            This option is using a potentially dangerous endpoint. Be careful when using that option, as roles
            are not restricting what can be issued anymore.

    encoding
        Can be either ``pem`` or ``der``. Defaults to ``pem``.

    exclude_cn_from_sans
        If set to true, Common name will not be part of the SANs.

    kwargs
        Any additional parameter accepted by Vault API or
        `x509_v2 module <https://docs.saltproject.io/en/latest/ref/modules/all/salt.modules.x509_v2.html#salt.modules.x509_v2.create_csr>`__
    """

    if csr is None and private_key is None:
        raise SaltInvocationError("either csr or private_key must be passed.")

    if csr is not None and private_key is not None:
        raise SaltInvocationError("only one of csr or private_key must be passed, not both")

    csr_args, extra_args = _split_csr_kwargs(kwargs)

    sign = "sign-verbatim" if sign_verbatim else "sign"

    endpoint = f"{mount}/{sign}/{role_name}"
    if issuer_ref is not None:
        endpoint = f"{mount}/issuer/{issuer_ref}/{sign}/{role_name}"

    payload = {k: v for k, v in extra_args.items() if not k.startswith("_")}

    payload["common_name"] = common_name

    if ttl is not None:
        payload["ttl"] = ttl

    payload["format"] = encoding
    payload["exclude_cn_from_sans"] = exclude_cn_from_sans

    if alt_names is not None:
        dns_sans, ip_sans, uri_sans, other_sans = _split_sans(alt_names)
        payload["alt_names"] = ",".join(dns_sans)
        payload["ip_sans"] = ",".join(ip_sans)
        payload["uri_sans"] = ",".join(uri_sans)
        payload["other_sans"] = ",".join(other_sans)

    # In case private_key is passed we're going to build
    # CSR in place.
    if private_key is not None:
        if isinstance(alt_names, dict):
            alt_names = [f"{k}:{v}" for k, v in alt_names.items()]

        if alt_names:
            csr_args["subjectAltName"] = alt_names

        csr_args["CN"] = common_name

        csr = __salt__["x509.create_csr"](
            private_key=private_key,
            private_key_passphrase=private_key_passphrase,
            digest=digest,
            **csr_args,
        )

    payload["csr"] = csr

    try:
        return vault.query("POST", endpoint, __opts__, __context__, payload=payload)["data"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def revoke_certificate(serial=None, certificate=None, mount="pki"):
    """
    Revoke issued certificate.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#revoke-certificate>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.revoke_certificate 7e:85:c5:d1:85:94:9a:46:08:b5:1b:9c:22:cb:35:e5:ea:f3:56:3f

    serial
        Specifies the serial of the certificate to revoke. Either ``serial`` or ``certificate`` must be specified.

    certificate
        Specifies the certificate (PEM or path) to revoke. Either ``serial`` or ``certificate`` must be specified.

        .. note::
            This parameter requires the :py:mod:`x509_v2 execution module <salt.modules.x509_v2>` to be available.

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """
    endpoint = f"{mount}/revoke/"
    payload = {}

    if serial is None and certificate is None:
        raise SaltInvocationError("either serial or certificate must be passed.")

    if serial is not None and certificate is not None:
        raise SaltInvocationError("only one of serial or certificate must be passed, not both")

    try:
        if certificate is not None:
            payload["certificate"] = __salt__["x509.encode_certificate"](
                certificate, encoding="pem"
            )
        elif serial is not None:
            if isinstance(serial, int):
                serial = dec2hex(serial)
            payload["serial_number"] = serial

        vault.query("POST", endpoint, __opts__, __context__, payload=payload, safe_to_retry=True)
        return True
    except vault.VaultInvocationError:
        return False
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def read_urls(mount="pki"):
    """
    Fetch the URLs to be encoded in generated certificates.
    No URL configuration will be returned until the configuration is set.

    `API method docs <https://developer.hashicorp.com/vault/api-docs/secret/pki#read-urls>`__.

    CLI Example:

    .. code-block:: bash

            salt '*' vault_pki.get_urls

    mount
        The mount path the PKI backend is mounted to. Defaults to ``pki``.
    """
    endpoint = f"{mount}/config/urls"

    try:
        return vault.query("GET", endpoint, __opts__, __context__)["data"]
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__}: {err}") from err


def _split_sans(sans) -> tuple[list, list, list, list]:
    dns_sans = []
    ip_sans = []
    uri_sans = []
    other_sans = []

    try:
        if isinstance(sans, list):
            sans = dict(map(lambda x: x.split(":", 1), sans))

        for k, v in sans.items():
            if k.upper() == "DNS" or k.upper() == "EMAIL":
                dns_sans.append(v)
            elif k.upper() == "IP":
                ip_sans.append(v)
            elif k.upper() == "URI":
                uri_sans.append(v)
            else:
                other_sans.append(f"{k};UTF8:{v}")
    except ValueError as err:
        raise CommandExecutionError(
            f"SAN is not in correct format. Must be in format <type>:<value>: {err}"
        ) from err

    return dns_sans, ip_sans, uri_sans, other_sans


def _split_csr_kwargs(kwargs):
    csr_args = {}
    extra_args = {}
    for k, v in kwargs.items():
        if k in VALID_CSR_ARGS:
            csr_args[k] = v
        else:
            extra_args[k] = v
    return csr_args, extra_args



import pytest

from tests.support.vault import vault_delete_secret
from tests.support.vault import vault_write_secret


@pytest.fixture(scope="class")
def vault_testing_values(container):  # pylint: disable=unused-argument
    vault_write_secret("secret/path/foo", success="yeehaaw")
    try:
        yield
    finally:
        vault_delete_secret("secret/path/foo")
